/*
 *  FreeMWare: run multiple x86 operating systems concurrently
 *  Copyright (C) 1999 The FreeMWare Team
 *
 *  emulation.c:  This file contains functions to emulate IA32 instructions
 *                
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

#include "host.h"
#include "freemware.h"
#include "monitor.h"

/*
 * Prototypes for instructions faulting in ring 3
 */
int emulate_hlt(void);
int emulate_cli(void);
int emulate_sti(void);
int emulate_ins(void);
int emulate_outs(void);
int emulate_clts(void);
int emulate_rdmsr(void);
int emulate_wrmsr(void);
int emulate_invd(void);
int emulate_wbinvd(void);
int emulate_invlpg(void);
int emulate_lgdt(u32 addr);
int emulate_lidt(u32 addr);
int emulate_lldt(u16 selector);
int emulate_ltr(u16 selector);
int emulate_lmsw(void);
int emulate_mov_from_cr(void);
int emulate_mov_to_cr(void);
int emulate_mov_from_dr(void);
int emulate_mov_to_dr(void);
int emulate_in(void);
int emulate_out(void);
int emulate_int3(void);
int emulate_load_ds(guest_context_t *context, u16 selector);
int emulate_load_es(guest_context_t *context, u16 selector);
int emulate_load_fs(guest_context_t *context, u16 selector);
int emulate_load_gs(guest_context_t *context, u16 selector);
int emulate_load_ss(guest_context_t *context, u16 selector);
int emulate_int(guest_context_t *context, int vector, u32 eip);
int emulate_iret(guest_context_t *context);


/*
 * Various utility routines to access guest memory
 *
 * FIXME -- all of these don't take segmentation or paging into account!
 */

static inline u8 read_guest_byte(u32 *addr)
{
    u8 *ptr = (u8 *)(*addr - vm_nexus()->mon_base);
    *addr += 1;
    return *ptr;
}
static inline u16 read_guest_word(u32 *addr)
{
    u16 *ptr = (u16 *)(*addr - vm_nexus()->mon_base);
    *addr += 2;
    return *ptr;
}
static inline u32 read_guest_dword(u32 *addr)
{
    u32 *ptr = (u32 *)(*addr - vm_nexus()->mon_base);
    *addr += 4;
    return *ptr;
}

static inline void write_guest_byte(u32 *addr, u8 val)
{
    u8 *ptr = (u8 *)(*addr - vm_nexus()->mon_base);
    *addr += 1;
    *ptr = val;
}
static inline void write_guest_word(u32 *addr, u16 val)
{
    u16 *ptr = (u16 *)(*addr - vm_nexus()->mon_base);
    *addr += 2;
    *ptr = val;
}
static inline void write_guest_dword(u32 *addr, u32 val)
{
    u32 *ptr = (u32 *)(*addr - vm_nexus()->mon_base);
    *addr += 4;
    *ptr = val;
}

static u32
guest_stack_pop(u32 *esp, int operand_size)
{
    switch (operand_size)
    {
    case 16: return read_guest_word(esp);
    case 32: return read_guest_dword(esp);
    }
    return 0;
}

static void
guest_stack_push(u32 *esp, int operand_size, u32 val)
{
    switch (operand_size)
    {
    case 16: *esp -= 2; *(u16 *)(*esp - vm_nexus()->mon_base) = val; break;
    case 32: *esp -= 4; *(u32 *)(*esp - vm_nexus()->mon_base) = val; break;
    }
}

static inline void
clear_descriptor(descriptor_t *descr)
{
    ((u32 *)descr)[0] = 0;
    ((u32 *)descr)[1] = 0;
}

static inline void
clear_gate(gate_t *gate)
{
    ((u32 *)gate)[0] = 0;
    ((u32 *)gate)[1] = 0;
}

static void 
read_guest_descriptor(descriptor_t *descr, u16 selector)
{
    if ( selector & 4 )
    {
        /* LDT selector -- not supported for now */

        clear_descriptor(descr);
    }
    else
    {
        /* GDT selector */

        u32 addr = selector & ~3;
        if (addr+7 > vm_nexus()->guest_gdt_info.limit) 
        {
            clear_descriptor(descr);
        }
        else
        {
            addr += vm_nexus()->guest_gdt_info.base;
            ((u32 *)descr)[0] = read_guest_dword(&addr);
            ((u32 *)descr)[1] = read_guest_dword(&addr);
        }
    }
}

static void 
write_guest_descriptor(descriptor_t *descr, u16 selector)
{
    if ( selector & 4 )
    {
        /* LDT selector -- not supported for now */
    }
    else
    {
        /* GDT selector */

        u32 addr = selector & ~3;
        if (addr+7 > vm_nexus()->guest_gdt_info.limit) 
            ;
        else
        {
            addr += vm_nexus()->guest_gdt_info.base;
            write_guest_dword(&addr, ((u32 *)descr)[0]);
            write_guest_dword(&addr, ((u32 *)descr)[1]);
        }
    }
}

static void 
read_guest_gate(gate_t *gate, int vector)
{
    u32 addr = 8*vector;
    if (   vector < 0 || vector >= 256 
        || addr+7 > vm_nexus()->guest_idt_info.limit) 
    {
        clear_gate(gate);
    }
    else
    {
        addr += vm_nexus()->guest_idt_info.base;
        ((u32 *)gate)[0] = read_guest_dword(&addr);
        ((u32 *)gate)[1] = read_guest_dword(&addr);
    }
}


/*
 * Operand decoding/evaluation routines
 */

#define FieldMod(modrm)     ( (modrm) >> 6       )
#define FieldReg(modrm)     (((modrm) >> 3) & 0x7)
#define FieldRM(modrm)      ( (modrm)       & 0x7)

#define FieldSIBScale(sib)  ( (sib) >> 6       )
#define FieldSIBIndex(sib)  (((sib) >> 3) & 0x7)
#define FieldSIBBase(sib)   ( (sib)       & 0x7)

static u32
decode_register(guest_context_t *context, int operand_size, int reg)
{
    switch ( operand_size )
    {
    case 8:
        switch ( reg )
        {
        case 0: return context->eax & 0xff;
        case 1: return context->ecx & 0xff;
        case 2: return context->edx & 0xff;
        case 3: return context->ebx & 0xff;
        case 4: return (context->eax >> 8) & 0xff;
        case 5: return (context->ecx >> 8) & 0xff;
        case 6: return (context->edx >> 8) & 0xff;
        case 7: return (context->ebx >> 8) & 0xff;
        }
        break;

    case 16:
        switch ( reg )
        {
        case 0: return context->eax & 0xffff;
        case 1: return context->ecx & 0xffff;
        case 2: return context->edx & 0xffff;
        case 3: return context->ebx & 0xffff;
        case 4: return context->esp & 0xffff;
        case 5: return context->ebp & 0xffff;
        case 6: return context->esi & 0xffff;
        case 7: return context->edi & 0xffff;
        }
        break;

    case 32:
        switch ( reg )
        {
        case 0: return context->eax;
        case 1: return context->ecx;
        case 2: return context->edx;
        case 3: return context->ebx;
        case 4: return context->esp;
        case 5: return context->ebp;
        case 6: return context->esi;
        case 7: return context->edi;
        }
        break;
    }
    return 0;
}

static u32
decode_memory_operand(guest_context_t *context, int operand_size, u32 addr)
{
    switch ( operand_size )
    {
    case 8:   return read_guest_byte(&addr);
    case 16:  return read_guest_word(&addr);
    case 32:  return read_guest_dword(&addr);
    }
    return 0;
}

static u32
decode_effective_address(guest_context_t *context, int address_size, 
                         u8 ModRegRM, u32 *addr)
{
    u32 mod = FieldMod(ModRegRM);
    if ( mod == 3 ) return 0;   /* not an effective address - can't happen */

    if ( address_size == 16 )
    { 
        u32 base = 0, displ = 0;

        /* Decode Base + Index */
        switch ( FieldRM(ModRegRM) )
        {
        case 0: base = context->ebx + context->esi; break;
        case 1: base = context->ebx + context->edi; break;
        case 2: base = context->ebp + context->esi; break;
        case 3: base = context->ebp + context->edi; break;
        case 4: base = context->esi; break;
        case 5: base = context->edi; break;
        case 6: base = mod? context->ebp : read_guest_word(addr); break;
        case 7: base = context->ebx;
        }

        /* Get Displacement */
        switch ( mod )
        {
        case 1:  displ = (signed char)read_guest_byte(addr); break;
        case 2:  displ = read_guest_word(addr); break;
        }

        return (base + displ) & 0xffff;
    }
    else
    { 
        u32 base, index = 0, displ = 0;

        /* Get SIB byte if present */
        if ( (base = FieldRM(ModRegRM)) == 4 )
        {
            u8 sib = read_guest_byte(addr);
            base = FieldSIBBase(sib);

            switch ( FieldSIBIndex(sib) )
            {
            case 0: index = context->eax; break;
            case 1: index = context->ecx; break;
            case 2: index = context->edx; break;
            case 3: index = context->ebx; break;
            case 4: index = 0;            break;
            case 5: index = context->ebp; break;
            case 6: index = context->esi; break;
            case 7: index = context->edi; break;
            }

            index <<= FieldSIBScale(sib);
        }

        /* Decode Base */
        switch ( base )
        {
        case 0: base = context->eax; break;
        case 1: base = context->ecx; break;
        case 2: base = context->edx; break;
        case 3: base = context->ebx; break;
        case 4: base = context->esp; break;
        case 5: base = mod? context->ebp : read_guest_dword(addr); break;
        case 6: base = context->esi; break;
        case 7: base = context->edi; break;
        }

        /* Get Displacement */
        switch ( mod )
        {
        case 1:  displ = (signed char)read_guest_byte(addr); break;
        case 2:  displ = read_guest_dword(addr); break;
        }

        return base + index + displ;
    }
}

static u32
decode_rm_operand(guest_context_t *context, int operand_size, int address_size,
                  u8 ModRegRM, u32 *addr)
{
    u32 ptr, val;

    if ( FieldMod(ModRegRM) == 3 )
        val = decode_register( context, operand_size, FieldRM(ModRegRM) );
    else
    {
        ptr = decode_effective_address( context, address_size, ModRegRM, addr );
        val = decode_memory_operand( context, operand_size, ptr );
    }

    return val;
}


int
emulate(guest_context_t *context)
/*
 *  emulation(): find out faulting instruction, jump to emulation function
 *
 *  INPUTS:
 *     none
 *
 *  OUTPUTS:
 *     0 if no success
 *     1 if success
 *     2 if success, but monitor needs to be remapped
 */ 
{
    int action = 0;
    int use32 = 1;       /* FIXME: is CS 16-bit or 32-bit */
    int address_size;
    int operand_size;
    int seg_override = 0;
    u8 c;
    u32 addr, val;
    u32 eip = context->eip;
    u32 esp = context->esp;
  
    address_size = use32? 32 : 16;
    operand_size = use32? 32 : 16;

    /*
     * Decode Intel instruction
     */

 prefix:
    c = read_guest_byte(&eip);
    switch (c) 
    {
    case 0x26: /* ES prefix */
        seg_override = context->es;
        goto prefix;
    case 0x2e: /* CS prefix */
        seg_override = context->cs;
        goto prefix;
    case 0x36: /* SS prefix */
        seg_override = context->ss;
        goto prefix;
    case 0x3e: /* DS prefix */
        seg_override = context->ds;
        goto prefix;
    case 0x64: /* FS prefix */
        seg_override = context->fs;
        goto prefix;
    case 0x65: /* GS prefix */
        seg_override = context->gs;
        goto prefix;
    case 0x66: /* Operand size prefix */
        operand_size = use32? 16 : 32;
        goto prefix;
    case 0x67: /* Address size prefix */
        address_size = use32? 16 : 32;
        goto prefix;

    case 0x07: /* pop ES */
        val = guest_stack_pop(&esp, operand_size);
        action = emulate_load_es(context, val);
        break;
    case 0x17: /* pop SS */
        val = guest_stack_pop(&esp, operand_size);
        action = emulate_load_ss(context, val);
        break;
    case 0x1f: /* pop DS */
        val = guest_stack_pop(&esp, operand_size);
        action = emulate_load_ds(context, val);
        break;

    case 0x8e: /* mov Sw, Ew */
        c = read_guest_byte(&eip);  /* RegModRM */
        val = decode_rm_operand(context, 16, address_size, c, &eip);
        switch ( FieldReg(c) )
        {
        case 0: action = emulate_load_es(context, val); break;
        case 1: /* mov CS, ... ??? */; break;
        case 2: action = emulate_load_ss(context, val); break;
        case 3: action = emulate_load_ds(context, val); break;
        case 4: action = emulate_load_fs(context, val); break;
        case 5: action = emulate_load_gs(context, val); break;
        }
        break;

    case 0xcd: /* int Ib */
        c = read_guest_byte(&eip);
        action = emulate_int(context, c, eip);
        break;

    case 0xcf: /* iret */
        action = emulate_iret(context);
        break;

    case 0xfa: /* cli */
        action = emulate_cli();
        break;

    case 0xfb: /* sti */
        action = emulate_sti();
        break;

    case 0x0f:
        /* there are some instructions beginning with 0x0f
         * we have a look on the second opcode byte
         */
        c = read_guest_byte(&eip);
        switch (c) 
        {
        case 0x00: /* Group 6 */
            c = read_guest_byte(&eip);
            switch (FieldReg(c))
            {
            case 2:  /* lldt */
                val = decode_rm_operand(context, 16, address_size, c, &eip);
                action = emulate_lldt(val);
                break;

            case 3:  /* ltr */
                val = decode_rm_operand(context, 16, address_size, c, &eip);
                action = emulate_ltr(val);
                break;
            }
            break;

        case 0x01: /* Group 7 */
            c = read_guest_byte(&eip);
            switch (FieldReg(c))
            {
            case 2:  /* lgdt */
                addr = decode_effective_address(context, address_size, c, &eip);
                action = emulate_lgdt(addr);
                break;

            case 3:  /* lidt */
                addr = decode_effective_address(context, address_size, c, &eip);
                action = emulate_lidt(addr);
                break;
            }
            break;

        case 0x06: /* clts */
            action = emulate_clts();
            break;

        case 0xa1: /* pop FS */
            val = guest_stack_pop(&esp, operand_size);
            action = emulate_load_fs(context, val);
            break;
        case 0xa9: /* pop GS */
            val = guest_stack_pop(&esp, operand_size);
            action = emulate_load_gs(context, val);
            break;
        }
        break;
    }


    /* Check next action */
    switch (action)
    {
    default:
    case 0:   /* Failure */
        return 0;

    case 1:   /* Retry instruction */
        return 1;

    case 2:   /* Skip instruction */
        context->eip = eip;
        context->esp = esp;
        return 1;

    case 3:   /* Remap monitor and skip instruction */
        context->eip = eip;
        context->esp = esp;
        return 2;
    }
}


int
emulate_cli(void)
     /*
      * Emulate cli instruction
      */
{
    /* here may reside some checks or so and emulation code */
    vm_nexus()->debug_msg.msg_code = EMU_CLI_MSG;
    return 0;
}

int
emulate_sti(void) 
     /* 
      * Emulate sti instruction
      */
{
    /* here may reside some checks or so and emulation code */
    vm_nexus()->debug_msg.msg_code = EMU_STI_MSG;
    return 0;
}

int
emulate_clts(void) 
     /* 
      * Emulate clts instruction
      */
{
    vm_nexus()->debug_msg.msg_code = EMU_CLTS_MSG;
    return 0;
}

int
emulate_lgdt(u32 addr) 
     /* 
      * Emulate lgdt instruction
      */
{
    descriptor_t *monitor_gdt = (descriptor_t *)vm_nexus()->gdt_offset;
    u16 limit = read_guest_word(&addr);
    u32 base  = read_guest_dword(&addr);

    int i, nr_descriptors;

    vm_nexus()->guest_gdt_info.limit = limit;
    vm_nexus()->guest_gdt_info.base  = base;

    if ( limit >= MON_GDT_PAGES*PAGE_SIZE ) return 0;
    nr_descriptors = (limit+1) / 8;

    for ( i = 1; i < nr_descriptors; i++ )
    {
        descriptor_t descr;
        read_guest_descriptor(&descr, i << 3);
        
        if ( descr.type & 0x10 )
        {
            /* Code or data segment descriptor */

            /* For now, we let everything go through unchanged, 
               only the DPL is forced to 3 */
            descr.dpl = D_DPL3;
        }
        else
        {
            /* System segment descriptor */
       
            switch ( descr.type )
            {
            case D_TSS: case D_TSS | 2:  /* available/busy TSS */
                /* Force the DPL to zero to disallow task switches */
                descr.dpl = D_DPL0;
                break;

            /* FIXME: Don't allow LDTs or gates for now ... */
            case D_LDT:
            case D_CALL:
            case D_INT: case D_TRAP:
            default:
                clear_descriptor(&descr);
                break;
            }
        }

        monitor_gdt[i] = descr;
    }
    for ( ; i < MON_GDT_PAGES*PAGE_SIZE / 8; i++ )
        clear_descriptor(monitor_gdt + i);

    /* We have just clobbered the monitor selectors.  Zero them out in the
       nexus data structure to reflect this.  See unmap_monitor() ... */
    vm_nexus()->mon_jmp_info.selector = 0;
    vm_nexus()->mon_stack_info.selector = 0;
    vm_nexus()->mon_tss_sel = 0;

    /* Return to the host to re-map the monitor */
    return 3;
}

int
emulate_lidt(u32 addr) 
     /* 
      * Emulate lidt instruction
      */
{
    u16 limit = read_guest_word(&addr);
    u32 base  = read_guest_dword(&addr);

    vm_nexus()->guest_idt_info.limit = limit;
    vm_nexus()->guest_idt_info.base  = base;

    return 2;
}

int
emulate_lldt(u16 selector) 
     /* 
      * Emulate lldt instruction
      */
{
    descriptor_t descr;

    if (selector & 4) return 0;    /* No LDT allowed */

    if (selector)
    {
        read_guest_descriptor(&descr, selector);
        if (!descr.p) return 0;
        if (descr.type != D_LDT) return 0;
    }
    else
    {
        clear_descriptor(&descr);
    }
    
    vm_nexus()->guest_ldt_sel = selector;
    vm_nexus()->guest_ldt_descr = descr;

    return 2;
}

int
emulate_ltr(u16 selector) 
     /* 
      * Emulate ltr instruction
      */
{
    descriptor_t descr;

    if (selector & 4) return 0;    /* No LDT allowed */

    read_guest_descriptor(&descr, selector);
    if (!descr.p) return 0;
    if (descr.type != D_TSS) return 0;
    
    descr.type |= 2;  /* set busy flag */

    write_guest_descriptor(&descr, selector);

    vm_nexus()->guest_tss_sel = selector;
    vm_nexus()->guest_tss_descr = descr;

    return 2;
}

int
emulate_load_ds(guest_context_t *context, u16 selector)
     /* 
      * Emulate instruction that loads DS register
      */
{
    vm_nexus()->debug_msg.msg_code = EMU_LOAD_SEGREG_MSG;
    vm_nexus()->debug_msg.para1 = 'D';
    vm_nexus()->debug_msg.para2 = selector;

    return 0;
}

int
emulate_load_es(guest_context_t *context, u16 selector)
     /* 
      * Emulate instruction that loads ES register
      */
{
    vm_nexus()->debug_msg.msg_code = EMU_LOAD_SEGREG_MSG;
    vm_nexus()->debug_msg.para1 = 'E';
    vm_nexus()->debug_msg.para2 = selector;

    return 0;
}

int
emulate_load_fs(guest_context_t *context, u16 selector)
     /* 
      * Emulate instruction that loads FS register
      */
{
    vm_nexus()->debug_msg.msg_code = EMU_LOAD_SEGREG_MSG;
    vm_nexus()->debug_msg.para1 = 'F';
    vm_nexus()->debug_msg.para2 = selector;

    return 0;
}

int
emulate_load_gs(guest_context_t *context, u16 selector)
     /* 
      * Emulate instruction that loads GS register
      */
{
    vm_nexus()->debug_msg.msg_code = EMU_LOAD_SEGREG_MSG;
    vm_nexus()->debug_msg.para1 = 'G';
    vm_nexus()->debug_msg.para2 = selector;

    return 0;
}

int
emulate_load_ss(guest_context_t *context, u16 selector)
     /* 
      * Emulate instruction that loads SS register
      */
{
    descriptor_t descr;

    if (!selector) return 0;
    if ((selector & 3) != vm_nexus()->guest_cpl) return 0;

    read_guest_descriptor(&descr, selector);
    if (!descr.p) return 0;
    if ((descr.type & 0x18) != D_DATA) return 0;
    if (!(descr.type & D_WRITE)) return 0;
    if (descr.dpl != vm_nexus()->guest_cpl) return 0;

    context->ss = selector | 3;

    return 2;
}


int
emulate_int(guest_context_t *context, int vector, u32 eip)
     /* 
      * Emulate INT instruction
      */
{
    u32 new_eip, new_esp;
    u16 new_cs, new_ss;

    gate_t gate;
    descriptor_t code_descr;
    descriptor_t stack_descr;

    int cpl = vm_nexus()->guest_cpl;

    /* Retrieve interrupt gate */
    read_guest_gate(&gate, vector);
    if (!gate.p) return 0;
    if (gate.type != D_INT && gate.type != D_TRAP) return 0;
    if (gate.dpl < cpl) return 0;
    if (!gate.selector) return 0;

    new_cs  = gate.selector;
    new_eip = (gate.offset_high << 16) | gate.offset_low;

    /* Retrieve code segment descriptor */
    read_guest_descriptor(&code_descr, new_cs);
    if (!code_descr.p) return 0;
    if ((code_descr.type & 0x18) != D_CODE) return 0;
    if (code_descr.dpl > cpl) return 0;
    if (new_eip > ((code_descr.limit_high << 16) | code_descr.limit_low)) return 0;

    if (!(code_descr.type & D_CONFORM) && code_descr.dpl < cpl)
    {
        /* Inter-privilege-level interrupt */
        descriptor_t *tss = &vm_nexus()->guest_tss_descr;
        u32 tss_limit = (tss->limit_high << 16) | tss->limit_low;
        u32 tss_base  = (tss->base_high << 24) | (tss->base_med << 16) | tss->base_low;

        u32 stack_addr = code_descr.dpl * 8 + 4;
        if (stack_addr+7 > tss_limit) return 0;
        stack_addr += tss_base;
        new_esp = read_guest_dword(&stack_addr);
        new_ss  = read_guest_dword(&stack_addr);

        read_guest_descriptor(&stack_descr, new_ss);
        if (!stack_descr.p) return 0;
        if ((stack_descr.type & 0x18) != D_DATA) return 0;
        if (!(stack_descr.type & D_WRITE)) return 0;
        if (stack_descr.dpl != code_descr.dpl) return 0;

        /* FIXME: stack limits check */

        /* Push old stack info */
        guest_stack_push(&new_esp, 32, (context->ss & ~3) | cpl);
        guest_stack_push(&new_esp, 32, context->esp);

        vm_nexus()->guest_cpl = code_descr.dpl;
    }
    else
    {
        /* Intra-privilege-level interrupt */

        new_ss  = context->ss;
        new_esp = context->esp;

        /* FIXME: stack limits check */
    }

    /* Push return info */
    guest_stack_push(&new_esp, 32, context->eflags);
    guest_stack_push(&new_esp, 32, (context->cs & ~3) | cpl);
    guest_stack_push(&new_esp, 32, eip);

    context->cs  = new_cs | 3;
    context->eip = new_eip;
    context->ss  = new_ss | 3;
    context->esp = new_esp;

    return 1;
}

int
emulate_iret(guest_context_t *context)
     /* 
      * Emulate IRET instruction
      */
{
    u32 esp = context->esp;

    u32 new_eip, new_esp, new_eflags;
    u16 new_cs, new_ss;

    descriptor_t code_descr;
    descriptor_t stack_descr;

    int cpl = vm_nexus()->guest_cpl;

    /* FIXME: stack limits check */

    new_eip    = guest_stack_pop(&esp, 32);
    new_cs     = guest_stack_pop(&esp, 32);
    new_eflags = guest_stack_pop(&esp, 32);

    if ((new_cs & 3) < cpl) return 0;
    if (!new_cs) return 0;
    read_guest_descriptor(&code_descr, new_cs);
    if (!code_descr.p) return 0;
    if ((code_descr.type & 0x18) != D_CODE) return 0;
    if (   (code_descr.type & D_CONFORM)
         ? (code_descr.dpl < cpl)
         : (code_descr.dpl != (new_cs & 3)) )  return 0;
    if (new_eip > ((code_descr.limit_high << 16) | code_descr.limit_low)) return 0;

    if ((new_cs & 3) > cpl)
    {
        /* Inter-privilege-level return */

        new_esp = guest_stack_pop(&esp, 32);
        new_ss  = guest_stack_pop(&esp, 32);

        if ((new_ss & 3) != (new_cs & 3)) return 0;
        if (!new_ss) return 0;
        read_guest_descriptor(&stack_descr, new_ss);
        if (!stack_descr.p) return 0;
        if ((stack_descr.type & 0x18) != D_DATA) return 0;
        if (!(stack_descr.type & D_WRITE)) return 0;
        if (stack_descr.dpl != (new_cs & 3)) return 0;

        /* FIXME: clear out invalid DS/ES/FS/GS selectors */

        vm_nexus()->guest_cpl = new_cs & 3;
    }
    else
    {
        /* Intra-privilege-level return */

        new_esp = esp;
        new_ss  = context->ss;
    }

    context->cs  = new_cs | 3;
    context->eip = new_eip;
    context->ss  = new_ss | 3;
    context->esp = new_esp;

    /* FIXME: set flags */

    return 1;
}

