/*
 *  FreeMWare: run multiple x86 operating systems concurrently
 *  Copyright (C) 1999  Kevin P. Lawton
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

#ifndef __MONITOR_H__
#define __MONITOR_H__

#include "descriptor.h"
#include "tss.h"
#include "paging.h"

#define RET_BECAUSE_REDIR     10
#define RET_BECAUSE_EXC       11
#define RET_BECAUSE_EMERR     12
#define RET_BECAUSE_MON_ERROR 13
#define RET_BECAUSE_REMAP     14

/*
 * Passing debug messages between host and guest context.
 */

typedef struct {
  // message code
  u32 msg_code;

  // parameter 1 and 2
  u32 para1, para2;
} debug_msg_t;

#define EMU_CLI_MSG 1
#define EMU_STI_MSG 2
#define EMU_CLTS_MSG 3
#define EMU_LGDT_MSG 4
#define EMU_LIDT_MSG 5
#define EMU_LLDT_MSG 6
#define EMU_LTR_MSG 7
#define EMU_LOAD_SEGREG_MSG 8



// Method1: push event info (CPU pushes error code before)
typedef struct {
  unsigned char pushl;    /* Always 0x68 == pushl            */
  u32  vector;            /* Interrupt vector number         */
  unsigned char jmp;      /* Always 0xe9 == jmp              */
  u32  reloc;             /* Relative offset of destination  */
  } __attribute__ ((packed)) idt_method1_t;

// Method2: push a dummy error first, then event info
typedef struct {
  unsigned char pushla;   /* Always 0x68 == pushl            */
  u32 dummy;              /* Dummy error code                */
  unsigned char pushlb;   /* Always 0x68 == pushl            */
  u32 vector;             /* Interrupt vector number         */
  unsigned char jmp;      /* Always 0xe9 == jmp              */
  u32 reloc;              /* Relative offset of destination  */
  } __attribute__ ((packed)) idt_method2_t;

typedef union {
  idt_method1_t m1;
  idt_method2_t m2;
  } idt_stub_t;


// I define the 'nexus' as the set of data structures which
// must exist in the current linear guest address space.  The
// host linear address space is not available while the current
// guest code is running, since we are using a completely
// different set of page mappings for the guest.  However,
// at some point an exception/interrupt will occur.  The
// interrupt mechanisms require that several structures exist in
// the current linear address space in order to service such
// an event.  These data structures make up part of our VM,
// a thin layer which exists in the guest.  Following is a
// list of what data structures compose this 'nexus':
//
//     - IDT (max  2048 bytes)
//     - GDT (max 65536 bytes)
//     - LDT (max 65536 bytes)
//     - TSS (max  8328 = 104 + 32 int redir + 8192 I/O permissions)
//     - kernel stack page
//     - transition code (host <--> guest)
//     - interrupt handler stubs
//     - Page Tables; PDE & PTE pages.

//
// Sizes of various nexus data structures used by the monitor
//

#define IDT_STUB_SIZE 15
#define BYTES2PAGES(b) ( ((b)+4095) >> 12 )

#define MON_IDT_SIZE       (8*256)
#define MON_GDT_SIZE       (8*6)
#define MON_LDT_SIZE       (8*1)
#define MON_IDT_STUBS_SIZE (IDT_STUB_SIZE*256)
#define MON_TSS_SIZE       (104)

#define MON_IDT_PAGES       BYTES2PAGES(MON_IDT_SIZE)
#define MON_GDT_PAGES       BYTES2PAGES(MON_GDT_SIZE)
#define MON_LDT_PAGES       BYTES2PAGES(MON_LDT_SIZE)
#define MON_IDT_STUBS_PAGES BYTES2PAGES(MON_IDT_STUBS_SIZE)
#define MON_TSS_PAGES       BYTES2PAGES(MON_TSS_SIZE)


/*
 *  Pages allocated for the VM by the host kernel driver.
 *  N Megs of physical memory are allocated, per the user's
 *  request, for the guest OS/application code.
 *  Additionally, some other overhead pages are allocated
 *  for structures such as the page directory, page tables,
 *  and other virtualized facilities.
 */

typedef struct {
    // requested size of the guest[] array in megs and pages
    unsigned guest_n_megs;
    unsigned guest_n_pages;

    // for the guest OS/app code
    u32 guest[FMW_MAX_PHY_MEGS * 256];

    // for the monitor's page directory
    u32 page_dir;

    // for the monitor's page table
    u32 page_tbl[(FMW_MAX_PHY_MEGS+3)>>2];

    // for the extra page table that maps our nexus code and structures
    u32 nexus_page_tbl;

    // We need a Page Table for identity mapping the transition code
    // between host and monitor spaces.
    u32 transition_PT;

    // Physical addresses of host pages which comprise the actual
    // monitor structures.  These will be mapped into the current
    // guest task's linear address space as well.
    u32  nexus;
    u32  idt[MON_IDT_PAGES];
    u32  gdt[MON_GDT_PAGES];
    u32  ldt[MON_LDT_PAGES];
    u32  tss[MON_TSS_PAGES];
    u32  idt_stubs[MON_IDT_STUBS_PAGES];
  } vm_pages_t;


typedef struct {
  // These fields are only used by the transition code.
  // They hold all info necessary to switch back to the host.
  gdt_info_t     host_gdt_info;
  gdt_info_t     host_idt_info;
  far_jmp_info_t host_jmp_info;
  far_jmp_info_t host_stack_info;
  u16            host_ldt_sel;
  u16            host_tss_sel;
  u32            host_cr0;
  u32            host_cr2;
  u32            host_cr3;
  u32            host_cr4;

  // These fields are filled by the host-side code, and used 
  // by the transition code.  They contain all info necessary
  // to switch to the monitor/guest address space.
  // This info changes whenever the monitor migrates.
  gdt_info_t     mon_gdt_info;
  gdt_info_t     mon_idt_info;
  far_jmp_info_t mon_jmp_info;
  far_jmp_info_t mon_stack_info;
  u16            mon_ldt_sel;
  u16            mon_tss_sel;
  u32            mon_cr3;
  u32            mon_base;

  // These fields contain info set by the guest
  gdt_info_t     guest_gdt_info;
  gdt_info_t     guest_idt_info;
  u16            guest_ldt_sel;
  u16            guest_tss_sel;
  descriptor_t   guest_ldt_descr;
  descriptor_t   guest_tss_descr;
  int            guest_cpl;

  // These fields contain info used by the transition code to
  // create the temporary identity mapping.  They never change.
  pageEntry_t    transition_pde;
  pageEntry_t   *transition_pde_p_host;
  pageEntry_t   *transition_pde_p_mon;
  u32            transition_laddr;

  // These fields contain the offsets relative to the monitor
  // data segment of various structures that make up the nexus.
  // These never change.
  u32 module_offset;
  u32 nexus_offset;
  u32 idt_offset;
  u32 gdt_offset;
  u32 ldt_offset;
  u32 tss_offset;
  u32 idt_stubs_offset;
  u32 page_dir_offset;

  // Debug messages returned from monitor-side fault handlers
  // to the host-side code.
  debug_msg_t    debug_msg;

  } __attribute__ ((packed)) nexus_t;



//
// Complete state of the VM (Virtual Machine)
//
typedef struct {
  vm_pages_t   pages; // memory pages allocated by the host

  struct {
    // Pointer to monitor IDT for use in host address space
    // Since the IDT is contained within 1 page, there's no
    // page span problem.
    gate_t       *idt;

    // Pointer to monitor TSS
    tss_t        *tss;

    // Pointer to monitor GDT, LDT
    descriptor_t *gdt;
    descriptor_t *ldt;

    idt_stub_t   *idt_stubs;

    pageEntry_t  *pageDirectory;

    // pointer into the monitor stack, so we can easily retrieve the
    // stack snapshot upon interrupt/exception.
    guest_context_t *guest_context;

    void (*__host2guest)(void);

    pageEntry_t   nexus_pde;
    } mon;

  nexus_t *nexus; // pointer to data structures in nexus

  unsigned initialized; // is the monitor initialized and OK?
  } vm_t;


extern char __nexus_start, __nexus_end, __mon_cs;
extern char __host2guest, __guest2host, __handle_fault;



/*
 * This structure describes the pages containing the code/data
 * of the monitor itself (inside the kernel module)
 */

#define FMW_MAX_MONITOR_PAGES 32

typedef struct {
    // virtual address space occupied by the module
    u32 start_addr;
    u32 end_addr;

    // number of pages
    unsigned n_pages;
    
    // the pages themselves
    u32 page[FMW_MAX_MONITOR_PAGES];
} monitor_pages_t;

extern monitor_pages_t monitor_pages;

/*
 * Access to nexus page of current VM from monitor space
 */

#define vm_nexus() ({           \
    u32 __dummy;                \
    __asm__ (                   \
        "movl %%esp,%0  \n"     \
        : "=r" (__dummy)        \
    );                          \
    (nexus_t *)(__dummy & ~0xfff);\
})


/*
 * Instruction emulation
 */

int emulate(guest_context_t *context);

int  init_monitor(vm_t *vm);
void host2guest(vm_t *vm);
void set_guest_context(vm_t *vm, guest_context_t *context);
void get_guest_context(vm_t *vm, guest_context_t *context);

#endif  // __GUEST_H__
