/*
 *  FreeMWare: run multiple x86 operating systems concurrently
 *  Copyright (C) 1999-2000  The FreeMWare developers team
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <signal.h>
#include <errno.h>
#include <stdarg.h>

#include "elf.h"

#include "freemware.h"
#include "user.h"

#include "decode.h"

#define COUNTER_OFF      0x08000

static void debug_exception( void *base, guest_context_t *context );
char *ptr = NULL;

/* Signal handler */
static int abort_vm = 0;
void abort_handler(int i)
{
    switch (i) {
	case SIGINT: printf("^C pressed. aborting execution\n");break;
	default: printf("Signal %d caught, aborting execution\n",i);
    }
    abort_vm = 1;
}

int
main(int argc, char *argv[])
{
    int fileno, virtno, ret;
    int request, data;
    unsigned i, timeout;
    struct stat statbuf;
    struct sigaction sg_act;
    guest_context_t context;

    int max_memory, config_fd, state, text_address, data_address, stack_address, bss_address;
    int debug_on, dump, is_elf;
    char c, config_file_name[256], guest_file_name[256], dump_file_name[256];
    char line[256];
    int line_index;
    void *elf_buf;
    Elf32_Ehdr eh;
    Elf32_Shdr sh, snst;
    unsigned int symstraddr;

    /* some inits */
    strcpy(config_file_name,"fmw.conf");
    guest_file_name[0]='\0';
    dump_file_name[0]='\0';
    timeout = 0;
    max_memory = 0;
    text_address = -1;
    data_address = -1;
    stack_address = -1;
    bss_address = -1;
    debug_on = 0;
    dump = 0;

    /* parsing command line for arguments */
    for(i=0;i<argc;i++) {
      if(argv[i][0]!='-') continue;
      switch(argv[i][1]) {
      case 'h':
	printf("FreeMWare Version, hm... don't know. Following options are for you:\n");
	printf("-h ... Print this helpful help\n");
	printf("-t ... Timeout (nr. of seconds the VM should run)\n");
	printf("-m ... Megs of memory for VM\n");
	printf("-f ... Config File (fully qualified path)\n");
	printf("-g ... Guest code (fully qualified path)\n");
	printf("-d ... Print debug messages\n");
	printf("-D ... Dump guest after running (human readable hex code)\n");
	exit(0);
	break;
      case 't':
	timeout = atoi(&argv[i][3]);
	break;
      case 'm':
	max_memory = atoi(&argv[i][3]);
	break;
      case 'f':
	strncpy(config_file_name,&argv[i][3],255);
	break;
      case 'g':
	strncpy(guest_file_name,&argv[i][3],255);
	break;
      case 'd':
	debug_on=1;
	break;
      case 'D':
	dump=1;
	break;
      default:
	fprintf(stderr,"Unknown option: %s\n",argv[i]);
	break;
      }
    }

    /* read config file */
    config_fd = open(config_file_name,O_RDONLY);
    if(config_fd == -1) {
      char error_msg[100];
      sprintf(error_msg,"error while opening %s",config_file_name);
      perror(error_msg);
      exit(1);
    }

    /* states
     * 0 reading
     * 1 ignoring
     */
    line_index = 0;

    while(read(config_fd,&c,1)) {
      line[line_index++]=c;
      if(c=='\n') {
	line[line_index-1]='\0';
	line_index = 0;
	if((line[0]=='#') || (line[0]=='\n')) continue; // is a comment or a blank line;
	while((line[line_index] != ' ') && (line[line_index] != '=')) line_index++;
	if(!strncmp(line,"timeout",line_index)) {
	  if(timeout != 0) {line_index=0;continue;} // command line overwrites config file settings;
	  while((line[line_index] == ' ') || (line[line_index] == '=')) line_index++;
	  timeout = atoi(&line[line_index]);
	} else if(!strncmp(line,"memory",line_index)) {
	  if(max_memory != 0) {line_index=0;continue;} // command line overwrites config file settings;
	  while((line[line_index] == ' ') || (line[line_index] == '=')) line_index++;
	  max_memory = atoi(&line[line_index]);
	} else if(!strncmp(line,"text_address",line_index)) {
	  if(text_address != -1) {line_index=0;continue;} // command line overwrites config file settings;
	  while((line[line_index] == ' ') || (line[line_index] == '=')) line_index++;
	  text_address = atoi(&line[line_index]);
	} else if(!strncmp(line,"data_address",line_index)) {
	  if(data_address != -1) {line_index=0;continue;} // command line overwrites config file settings;
	  while((line[line_index] == ' ') || (line[line_index] == '=')) line_index++;
	  data_address = atoi(&line[line_index]);
	} else if(!strncmp(line,"stack_address",line_index)) {
	  if(stack_address != -1) {line_index=0;continue;} // command line overwrites config file settings;
	  while((line[line_index] == ' ') || (line[line_index] == '=')) line_index++;
	  stack_address = atoi(&line[line_index]);
	} else if(!strncmp(line,"bss_address",line_index)) {
	  if(bss_address != -1) {line_index=0;continue;} // command line overwrites config file settings;
	  while((line[line_index] == ' ') || (line[line_index] == '=')) line_index++;
	  bss_address = atoi(&line[line_index]);
	} else if(!strncmp(line,"guest",line_index)) {
	  if (guest_file_name[0] != '\0') {line_index=0;continue;}
	  while((line[line_index] == ' ') || (line[line_index] == '=')) line_index++;
	  strcpy(guest_file_name,&line[line_index]);
	}
	line_index = 0;
      }
    }

    /* end of file */

    close(config_fd);

    if (max_memory == 0)
      max_memory = FMW_MAX_PHY_MEGS;
    if (guest_file_name[0] == '\0')
      strcpy(guest_file_name,"../guest/virtcode/virtcode.bin");
    if (dump_file_name[0] == '\0')
      strcpy(dump_file_name,"dumpfile");

    if(debug_on) {
      fprintf(stderr,"Timeout: %d\n",timeout);
      fprintf(stderr,"Memory: %d\n",max_memory);
      fprintf(stderr,"Text address: %d\n",text_address);
      fprintf(stderr,"Data address: %d\n",data_address);
      fprintf(stderr,"Stack address: %d\n",stack_address);
      fprintf(stderr,"BSS address: %d\n",bss_address);
    }

    fprintf(stderr, "Loading guest code: %s\n", guest_file_name);
    virtno = open(guest_file_name, O_RDONLY);
    if (virtno < 0) {
        perror("open");
        exit(1);
    }
    if (fstat(virtno, &statbuf) != 0) {
        perror("fstat");
        exit(1);
    }
    fprintf(stderr,"Allocating 0x%x bytes of memory for reading file\n",statbuf.st_size);
    if((elf_buf=(void *)malloc(statbuf.st_size)) == NULL) {
      perror("malloc");
      exit(1);
    }
    
    if (read(virtno, elf_buf, statbuf.st_size) != statbuf.st_size) {
        perror("read");
        exit(1);
    }
    close(virtno);

    if((((char *)elf_buf)[0]!=0x7f) ||
       (((char *)elf_buf)[1]!='E') ||
       (((char *)elf_buf)[2]!='L') ||
       (((char *)elf_buf)[3]!='F')) {
      fprintf(stderr,"This is no ELF binary, hope this is a pure binary an run it as is\n");
      is_elf = 0;
    }
    else {
      printf("ok, this is an ELF-Binary\n");
      is_elf = 1;
      eh=((Elf32_Ehdr *)elf_buf)[0];
      if(eh.e_type!=ET_EXEC) {
	fprintf(stderr,"Please, try it with an executable!\n");
	exit(1);
      }
      if(eh.e_machine!=EM_386) {
	fprintf(stderr,"FreeMWare isn't able to run binaries other than such for x86.\nPlease use an emulator for your binary!");
	exit(1);
      }
      snst=(((Elf32_Shdr *)(elf_buf+eh.e_shoff))[eh.e_shstrndx]);
      symstraddr=snst.sh_offset;

      for(i=0;i<eh.e_shnum;i++) {
	sh=(((Elf32_Shdr *)(elf_buf+eh.e_shoff))[i]);
	if(sh.sh_name && !strcmp((char *)elf_buf+symstraddr+sh.sh_name,".text"))
          text_address = sh.sh_addr;
      }
    }


    fprintf(stderr, "Opening VM\n");
    fileno = open("/dev/freemware0", O_RDWR);
    if (fileno < 0) {
        fprintf(stderr, "error: open returned error\n");
        exit(1);
    }

    fprintf(stderr, "Allocating %dMB of physical memory in VM\n",max_memory);
    request = FMWALLOCVPHYS;
    data = max_memory;
    ret = ioctl(fileno, request, data);
    if (ret < 0) {
        fprintf(stderr, "error: ioctl failed\n");
        goto close_fd;
    }

    fprintf(stderr, "Mapping virtualized physical memory into user monitor\n");
    ptr = mmap(NULL, max_memory*1024*1024, PROT_READ|PROT_WRITE, MAP_SHARED, fileno, 0);
    if (ptr == (void *)-1) 
        perror("mmap");

    fprintf(stderr, "Zeroing virtualized physical memory\n");
    memset(ptr, 0, max_memory*1024*1024);

    fprintf(stderr, "Loading guest code to virtualized physical memory\n");
    if (!is_elf)
      memcpy(ptr+text_address,elf_buf,statbuf.st_size);
    else
    {
      eh=((Elf32_Ehdr *)elf_buf)[0];

      for(i=0;i<eh.e_shnum;i++) {
	sh=(((Elf32_Shdr *)(elf_buf+eh.e_shoff))[i]);
	if((sh.sh_flags & SHF_ALLOC) && (sh.sh_type!=SHT_NOBITS)) {
	  /* load if it occupies spaces and has bits */
	  memcpy(ptr+sh.sh_addr,elf_buf+sh.sh_offset,sh.sh_size);
	}
      }
    }
    free(elf_buf);

    fprintf(stderr, "Setting up timeout after %d seconds\n", timeout);

    memset(&sg_act, 0, sizeof(sg_act));
    sg_act.sa_handler = abort_handler;
    sg_act.sa_flags = SA_RESETHAND;
    sigaction(SIGINT,&sg_act,NULL);

    // Set up timers and IO device emulation for user space
    // monitor.
    io_init();

    memset( &context, 0, sizeof(context) );
    context.eip = text_address;
    context.esp = stack_address;

    fprintf(stderr, "Running VM\n");
    request = FMWRUNGUEST;
    data = (unsigned long)&context;
    do {
        ret = ioctl(fileno, request, data);
    } while (!abort_vm && ret < 0 && errno == EINTR);
    if (ret < 0 && errno != EINTR) {
        perror("ioctl");
        goto close_fd;
    }

    if (!abort_vm)
        debug_exception( ptr, &context );

    fprintf(stderr, "Guest counter: %d\n", *(int *)(ptr + COUNTER_OFF));

    if(dump) {
      int dumpno, counter, tmp;
      char outbuf;

      counter = 0;
      fprintf(stderr,"Dumping memory to %s\n",dump_file_name);
      dumpno = open(dump_file_name, O_CREAT|O_RDWR,S_IRUSR|S_IWUSR);
      for(i=0;i<max_memory*1024*1024;i++) {
	tmp=((char *)ptr)[i]>>4;
	if(tmp > 9)
	  tmp+=(int)'A';
	else
	  tmp+=(int)'0';
	outbuf=(char)tmp;
	write(dumpno,&outbuf,1);
	tmp=((char *)ptr)[i] & 0x0f;
	if(tmp > 9)
	  tmp+=(int)'A';
	else
	  tmp+=(int)'0';
	outbuf=(char)tmp;
	write(dumpno,&outbuf,1);

	counter++;
	if(!(counter%4)) {
	  outbuf=' ';
	  write(dumpno,&outbuf,1);
	}
	if(!(counter%16)) {
	  outbuf='\n';
	  write(dumpno,&outbuf,1);
	}
      }
      close(dumpno);
    }

    fprintf(stderr, "Unmapping guest physical memory\n");
    if (munmap(ptr, max_memory*1024*1024) != 0) {
        perror("munmap");
        exit(1);
    }

    fprintf(stderr, "Tearing down VM\n");
    request = FMWTEARDOWN;
    data = 0;
    ret = ioctl(fileno, request, data);
    if (ret < 0) {
        fprintf(stderr, "error: ioctl failed\n");
    }

close_fd:
    fprintf(stderr, "Closing VM\n");
    ret = close(fileno);
    if (ret < 0) {
        fprintf(stderr, "error: close failed\n");
        exit(1);
    }

    exit(0);
}

static void debug_exception( void *base, guest_context_t *context )
{
    struct i386_context ctx;
    struct i386_decode instr;
    int i, len = 8;
    char prefix[256], text[256];

    memset( &ctx, 0, sizeof(ctx) );
    memset( &instr, 0, sizeof(instr) );

    ctx.mode    = CODE32;      /* FIXME */
    ctx.base    = base;
    ctx.segment = context->cs;
    ctx.offset  = context->eip;

    if ( !i386_decode( &ctx, &instr ) || !i386_decode_text( &ctx, &instr, text ) )
        strcpy( text, "???" );
    else
        len = instr.length;

    sprintf( prefix, "%04lX.%08lX  ", context->cs, context->eip );

    for ( i = 0; i < len && i < 12; i++ )
        sprintf( prefix+strlen(prefix), "%02X", ((unsigned char *)base)[ context->eip + i ] );
    for ( ; i < 12; i++ )
        strcat( prefix, "  " );


    fprintf(stderr, "\n");
    fprintf(stderr, "Abort due to exception %d at %04lx:%08lx\n", 
                    context->event_info&0xff, context->cs, context->eip );
    fprintf(stderr, "\n");

    fprintf(stderr, "Register dump:\n");
    fprintf(stderr, " CS:%04x SS:%04x DS:%04x ES:%04x FS:%04x GS:%04x\n",
                    context->cs, context->ss, 
                    context->ds, context->es, context->fs, context->gs );
    fprintf(stderr, " EAX:%08lx EBX:%08lx ECX:%08lx EDX:%08lx\n",
                    context->eax, context->ebx, context->ecx, context->edx );
    fprintf(stderr, " ESI:%08lx EDI:%08lx EBP:%08lx ESP:%08lx EFLAGS:%08lx\n",
                    context->esi, context->edi, context->ebp, context->esp, context->eflags );
    fprintf(stderr, "\n");

    fprintf(stderr, "Stack dump:\n");
    for ( i = 0; i < 4; i++ )
        fprintf(stderr, " %08lx:  %08lx %08lx %08lx %08lx\n",
                        context->esp + 16*i,
                        *(unsigned long *)(base + context->esp + 16*i),
                        *(unsigned long *)(base + context->esp + 16*i + 4),
                        *(unsigned long *)(base + context->esp + 16*i + 8),
                        *(unsigned long *)(base + context->esp + 16*i + 12));
    fprintf(stderr, "\n");

    fprintf(stderr, "Current instruction:\n" );
    fprintf(stderr, " %s %s\n", prefix, text );
    fprintf(stderr, "\n");
}

  void
panic(char *fmt, ...)
{
  va_list ap;

#if 0
  if (bx_logfd) {
    fprintf(bx_logfd, "panic, ");

    va_start(ap, fmt);
    vfprintf(bx_logfd, fmt, ap);
    va_end(ap);
    }
#endif
  fprintf(stderr, "panic, ");

  va_start(ap, fmt);
  vfprintf(stderr, fmt, ap);
  va_end(ap);

  exit(-1);
}
