/*    
	WinConfig.c	1.23
    	Copyright 1997 Willows Software, Inc. 

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; see the file COPYING.LIB.  If
not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA.


For more information about the Willows Twin Libraries.

	http://www.willows.com	

To send email to the maintainer of the Willows Twin Libraries.

	mailto:twin@willows.com 
 */

/*************************************************************************
 *
 * Manage configuration parameters through a single configuration file.
 * 
 * These environment variables control parameters:
 *
 * TWINRC		location of configuration file if set
 * 			default, current working directory, home, willows
 * HOME			location of users home directory if set
 * 			default, value retrieved from pwnam
 * LOGNAME		user name if set
 *			default, use value from pwnam
 * DISPLAY		display server if set
 * 			default, value retrieved from twinrc file
 * PATH			path used to find and open files,
 *			default if not in TWINRC is to use PATH
 *			
 *************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "Driver.h"
#include "WinConfig.h"
#include "platform.h"
#include "compat.h"

/**************************************************************************
 *
 * [boot]		startup configuration settings
 * [windows]		run-time configuration settings
 * [xdos]		mapping of DOS drives
 * [commondialogs]	how to drive common dialog fileopen dialog
 *
 * [fontalias]		what font mapping to use
 * [colors]		what colors to use
 *  
 **************************************************************************
 *
 *	parameter	what parameter is being addressed
 *	opcode	
 *		1	use environment as override if set
 *			use section, keyname, default to lookup string
 *		6	use environment as override if set
 *			use section, keyname, default to lookup string
 *			if null, return path to TWINRC
 *
 *	lpszsection	section to use
 *	lpszkeyname	keyname to use
 *	lpszdefault	default value
 *	lpszenviron	environment string
 *	lpszstring	allocated buffer
 *
 *************************************************************************/


static char *sections[] = {
	"boot",
	"windows",
	"XDOS",
	"CommonDialogs"
};

static TWIN_CONFIG  *wcpdata = NULL;


/* name of configuration file */
extern BOOL TWIN_InitDriver();

BOOL InitTwinConfig();

char *
GetTwinFilename(void)
{
	static char *twinrc;

	/* have we already set it */
	if(twinrc)
		return twinrc;

	TWIN_InitDriver();
	twinrc = (char *) DRVCALL_CONFIG(PCFGH_GETFILENAME, 0, 0 , NULL);

	return twinrc;
	
}


char *
GetTwinBuffer(TWIN_CONFIG *tp,char *wcpbuf)
{
	/* if we have a buffer, use it */
	if(wcpbuf)
		return wcpbuf;

	/* if we already allocated a buffer, return it */
	if(tp->lpszstring)
		return tp->lpszstring;

	/* allocate a permanent buffer */
	tp->lpszstring = (char *) WinMalloc(256);

	/* return it */
	return tp->lpszstring;	
}

char *
GetTwinString(int wcp, char *wcpbuf,int nsize )
{
	TWIN_CONFIG  *tp;
	char     *p = NULL;
	static char *twinrc;

	InitTwinConfig();

	if(twinrc == 0)
		twinrc = GetTwinFilename();
	
	if (wcpdata == NULL)
		wcpdata = (TWIN_CONFIG *)DRVCALL_CONFIG(PCFGH_GETDEFAULTS, 0, 0 , NULL);
	for(tp = wcpdata;tp->parameter > 0;tp++) {
		if(tp->parameter == wcp) {
		  switch(tp->opcode) {
			case 1:
				if(tp->lpszenviron && 
				   (p = (char *) getenv(tp->lpszenviron))) {
				       strcpy(GetTwinBuffer(tp,wcpbuf),p);

				       /* we got the environment variable */
				       return GetTwinBuffer(tp,wcpbuf);
				} 
				p = GetTwinBuffer(tp,wcpbuf);
				if(nsize == 0)
					nsize = 256;
				GetPrivateProfileString(
					sections[tp->section],
					tp->lpszkeyname,
					tp->lpszdefault,
					p,
					nsize, 
					twinrc);

				return GetTwinBuffer(tp,wcpbuf);

			 case 6:
				if(tp->lpszenviron && 
				   (p = (char *) getenv(tp->lpszenviron))) {
				       strcpy(GetTwinBuffer(tp,wcpbuf),p);

				       /* we got the environment variable */
				       return GetTwinBuffer(tp,wcpbuf);
				}
			
				p = GetTwinBuffer(tp,wcpbuf);
				if(nsize == 0)
					nsize = 256;
				/* get the value from the file */
				GetPrivateProfileString(
		  			sections[tp->section],
					tp->lpszkeyname,
					tp->lpszdefault,
					p,
	    				nsize, 
					twinrc);

				/* if we got something return it */
				if(strlen(p)) {
					return p;
				}
					
				/* use dirname(TwinFilename) */
				strcpy(p, GetTwinFilename());

				/* worst case same as TwinFilename directory */
				return p;
				
			default:
				break;
		  }
		  return wcpbuf;
		}
	}

	return 0;
}

unsigned long 
GetTwinInt(int wcp)
{
	char wcpbuffer[256];
	char *wcptr;

	wcptr = GetTwinString(wcp,wcpbuffer,256);
	
	return strtol(wcptr,0,0);
}

typedef	struct {
	int	value;
	char   *string;
} CONFIGOPTION;

static CONFIGOPTION configoption[] = {
	{ 0, "debugger"},	/* load the the debugger 		*/
	{ 0, "noditherpure"},	/* how to handle dithering      	*/
	{ 0, "colormap"},	/* allocate private colormap    	*/
	{ 0, "hiresfonts"},	/* use hi-res fonts, not low 		*/
	{ 0, "vgamode"}, 	/* default to 640x480x16colors  	*/
	{ 0, "nothreads"},	/* don't allow threads			*/
	{ 0, "fatalerr"},	/* break into debugger 			*/
	{ 0, "handlerr"},	/* break on handleobj errors 		*/
	{ 0, "noapistubs"},	/* NO msgbox for STUB 			*/
	{ 0, "nopaging"},  	/* disable demand page loading  	*/
	{ 0, "nochangecase"},	/* don't change case of filenames 	*/
	{ 0, "mapcurdir"},	/* xdos map current directory   	*/
	{ 0, "colormask"},	/* use rgb colormask to convert mono    */

	{ 0, "failerr"},	/* break on log FAIL errors 		*/
	{ 0, "warnerr"},	/* break on log WARN errors 		*/
	{ 0, "fixmeerr"},	/* break on log FIXME errors 		*/
	{ 0, "stuberr"},	/* break on log STUB  errors 		*/
	
};	

int
GetConfigOption(int uid)
{
	char  *twinrc = GetTwinFilename();
	char   *p = (char *) 0;
	int    dv = 0;
	int    rc = 0;

	switch(uid) {
	case  WD_DEBUGGER:      /* load the the debugger 		*/
		dv = configoption[WD_DEBUGGER].value;
		p  = configoption[WD_DEBUGGER].string;
#ifdef DEBUG
		dv = 1;
#endif
		break;
	case  WD_NODITHERPURE: 	/* how to handle dithering      	*/
		dv = configoption[WD_NODITHERPURE].value;
		p  = configoption[WD_NODITHERPURE].string;
		break;
	case  WD_COLORMAP:	/* allocate private colormap    	*/
		dv = configoption[WD_COLORMAP].value;
		p  = configoption[WD_COLORMAP].string;
		break;
	case  WD_HIRESFONTS: 	/* use hi-res fonts, not low 		*/
		dv = configoption[WD_HIRESFONTS].value;
		p  = configoption[WD_HIRESFONTS].string;
		break;
	case  WD_VGAMODE:	/* default to 640x480x16colors  	*/
		dv = configoption[WD_VGAMODE].value;
		p  = configoption[WD_VGAMODE].string;
		break;
	case  WD_NOTHREADS:	/* disable thread library calls		*/
		dv = configoption[WD_NOTHREADS].value;
		p  = configoption[WD_NOTHREADS].string;
		break;
	case  WD_FATALERR:     	/* break into debugger 			*/
		dv = configoption[WD_FATALERR].value;
		p  = configoption[WD_FATALERR].string;
		break;
	case  WD_HANDLERR:	/* break on handleobj errors 		*/
		dv = configoption[WD_HANDLERR].value;
		p  = configoption[WD_HANDLERR].string;
		break;
	case  WD_APISTUBS:     	/* break, NO msgbox for STUB 		*/
		dv = configoption[WD_APISTUBS].value;
		p  = configoption[WD_APISTUBS].string;
		break;
	case  WD_NOPAGING:	/* disable demand page loading  	*/
		dv = configoption[WD_NOPAGING].value;
		p  = configoption[WD_NOPAGING].string;
		break;
	case  WD_NOCHANGECASE:  /* don't change case of filenames 	*/
		dv = configoption[WD_NOCHANGECASE].value;
		p  = configoption[WD_NOCHANGECASE].string;
		break;
	case  WD_MAPCURDIR:	/* xdos map current directory   	*/
		dv = configoption[WD_MAPCURDIR].value;
		p  = configoption[WD_MAPCURDIR].string;
		break;
	case  WD_COLORMASK:	/* use rgb masks to conver mono colors   */
		dv = configoption[WD_COLORMASK].value;
		p  = configoption[WD_COLORMASK].string;
		break;
	/*****************************************************************/

	case  WD_FAILERR:	/* use rgb masks to conver mono colors   */
		dv = configoption[WD_FAILERR].value;
		p  = configoption[WD_FAILERR].string;
		break;
	case  WD_WARNERR:	/* use rgb masks to conver mono colors   */
		dv = configoption[WD_WARNERR].value;
		p  = configoption[WD_WARNERR].string;
		break;
	case  WD_FIXERR:	/* use rgb masks to conver mono colors   */
		dv = configoption[WD_FIXERR].value;
		p  = configoption[WD_FIXERR].string;
		break;
	case  WD_STUBERR:	/* use rgb masks to conver mono colors   */
		dv = configoption[WD_STUBERR].value;
		p  = configoption[WD_STUBERR].string;
		break;
	default:
		return 0;
	}
	
	rc = GetPrivateProfileInt("boot",p,dv,twinrc);	

	return rc;
}

/********************************************************************
*	GetEnv
*
*	Our version of getenv that reads the [Environment] section of the twinrc file to obtain variables.
********************************************************************/
char *GetEnv(const char *env)
{
	char *twinrc = GetTwinFilename();
	static char buf[256];
	int count;

	/* Try and retrieve the environment variable */
	count = GetPrivateProfileString(sections[WCS_ENV], env, NULL, buf, 256,
		twinrc);

	if (count == 0)
		/* The default was used, meaning nothing was found */
		return(NULL);
	else
		/* Return the buffer containing the value found in the twinrc
		 * [Environment] section.
		 */
		return(buf);
}

BOOL InitTwinConfig()
{
    static BOOL binit = FALSE;

    if (!binit) {
     FILE*           fp;
     TWIN_CONFIG  *tp;
     char *twinrc;
     char buf[MAX_PATH];

     binit = TRUE;

     twinrc = GetTwinFilename();

     if (wcpdata == NULL)
          wcpdata = (TWIN_CONFIG *)DRVCALL_CONFIG(PCFGH_GETDEFAULTS, 0, 0 ,
NULL);

     /* Try to open .twinrc */
     fp = fopen(twinrc, "r");

     if (!fp) { /* Failed : create default with only boot section */
         fp = fopen(twinrc, "w");
         fprintf(fp,"%s\n","[boot]");

         for(tp = wcpdata;tp->parameter > 0;tp++) {
          if ( tp->parameter && tp->section == WCS_BOOT &&
               tp->lpszkeyname && *(tp->lpszkeyname) ) {
              strcpy(buf,tp->lpszkeyname);
              strcat(buf,"=");
              strcat(buf,tp->lpszdefault ? tp->lpszdefault : "");

              fprintf(fp,"%s\n",buf);
          }
         }
     }

     fclose(fp);
    } /* if (!binit) */

   return TRUE;
}
