#!/usr/bin/perl
#
# Copyright (C) 1999 VMware, Inc.  All Rights Reserved.
#
# BSD guest OS tools installation
#
# If your copy of perl is not in /usr/bin, please adjust the line above.


die "You must be root to install the VMware Guest OS Tools.\n" unless $> == 0;

print << "EOM";
---------------------------------------
    VMware Guest OS Tools Installer
    Copyright (C) 1999 VMware, Inc.
---------------------------------------
EOM

my ($default_initdir, $xconfig_file, $xserver_link);
my ($screen_mode, $modules_src_file);


#
# makeDirectory(dir, logging)
#
# Make the directory, if it doesn't exist
#
sub makeDirectory {
    my ($path, $logging) = @_;
    my $slashpos;
    my $head;

    if ("$path" eq "") {
	return;
    }

    if (-d $path) {
	return;
    }

    if (-e "$path") {
	die "\n### $path exists but it is not a directory.\n";
    } else {
	$slashpos = rindex($path, "/");
	$head = substr($path, 0, $slashpos);
	&makeDirectory($head, $logging);
	if ($logging) {
	    &Log("directory $path\n");
	}
	mkdir($path, 0755) or die "\n### Cannot create $path: $!.\n";
    }
}


#
# LogFile(filename)
#
# Write filename and modification time to log
#
sub LogFile {
    my $filename = shift;
    my ($dev, $ino, $mode, $nlink, $uid, $gid, $rdev, $size,
	      $atime, $mtime, $ctime) = stat($filename);
    &Log("file $filename $mtime\n");
}


#
# Log(message)
#
# Write message to log
#
sub Log {
  my $string = shift;
  print LINUX_TOOLS_LOG $string;
}


#
# OpenLog
#
# Open the log file, creating it if necessary
#
sub OpenLog {
  &makeDirectory("/etc/vmware", 0);
  open(LINUX_TOOLS_LOG,">/etc/vmware/tools_log") or
     die "\n### Cannot open /etc/vmware/tools_log\n";
}


#
# copydist(mode, src, dest)
#
# Copy the file and set its mode.  If the file has a .dist extension,
# perform keyword substitution first.
#
sub copydist {
  my ($mode, $src, $dst) = @_;

  if (-e $dst) {
      die "\n### File $dst should not exist.\n" .
	  "### Please remove it and try again.\n";
  }

  open(SRC, "<$src") or die "\n### open: $src: $!.\n";
  open(DST, ">$dst") or die "\n### open: $dst: $!.\n";
  my $command = "chmod " . $mode . " " . $dst;
  if (system($command) != 0) {
    print "The command\n  " . $command . "\n failed: $!.\n";
    exit(1);
  }

  if ($src =~ /\.dist$/) {              # Need to make changes to contents
    while(defined($_ = <SRC>)) {
      s/%XCONFIG_FILE%/$xconfig_file/g;
      s/%XSERVER_LINK%/$xserver_link/g;
      s/%XSERVER_BIN%/$xserver_bin/g;
	s/%SCREEN_MODE%/$screen_mode/g;
      print DST $_;
    }
  } else {                              # Just copy
    while(defined($_ = <SRC>)) {
      print DST $_;
    }
  }

  close(SRC);
  close(DST);
}


#
# install(action, mode, src, dst, back)
#
# Install the file in the destination with the given mode.
# Action is one of link/copy/dist.  Dist is like copy, but
# with keyword substitution.
#
sub install {
    my ($action, $mode, $src, $dst, $back) = @_;

    if (-e $dst) {
      if ($back) {
	  if (-e $dst . ".old.3") {
	      system ("rm -f $dst" . ".old.3");
	  }
	TRY:
	  for $i (1..99) {
	      if (! -e $dst . ".old." . $i) {
		  rename($dst, $dst . ".old." . $i);
		  last TRY;
	      }
	  }
       } else {
	  system ("rm -f $dst");
       }
  }

    my $path = substr($dst, 0, rindex($dst, "/"));
    &makeDirectory($path, 1);

    if ($action eq "dist") {
	# Distribute action
	&copydist($mode, $src, $dst);
	&LogFile($dst);
    } elsif ($action eq "copy") {
	# Copy action
	my $command = "cp " . $src . " " . $dst;
	if (system($command) != 0) {
	    print "The command\n  " . $command . "\n failed: $!.\n";
	    exit(1);
	}
	$command = "chmod " . $mode . " " . $dst;
	if (system($command) != 0) {
	    print "The command\n  " . $command . "\n failed: $!.\n";
	    exit(1);
	}
	&LogFile($dst);
    } elsif ($action eq "move") {
	# Move action
	my $command = "mv " . $src . " " . $dst;
	if (system($command) != 0) {
	    print "The command\n  " . $command . "\n failed: $!.\n";
	    exit(1);
	}
	&Log("move $src $dst\n");
    } else {
	# Link action
	my $command = "ln -s " . $src . " " . $dst;
	if (system($command) != 0) {
	    print "The command\n  " . $command . "\n failed: $!.\n";
	    exit(1);
	}
	&Log("link $src $dst\n");
    }
}


#
# pathCompare(dir, path1, path2)
#
# Compare the two paths, and return true if they are identical
# Evaluate the paths with respect to the passed in directory
#
sub pathCompare {
    my ($dir, $path1, $path2) = @_;

    # Prepend directory for relative paths
    $path1 =~ s|^([^/])|$dir/$1|;
    $path2 =~ s|^([^/])|$dir/$1|;

    # Squash out ..'s in paths
    while ($path1 =~ /\/.*\/\.\.\//) {
	$path1 =~ s|/[^/]*/\.\./|/|;
    }

    while ($path2 =~ /\/.*\/\.\.\//) {
	$path2 =~ s|/[^/]*/\.\./|/|;
    }

    # Squash out .'s in paths
    while ($path1 =~ /\/\.\//) {
	$path1 =~ s|/\./|/|;
    }

    while ($path2 =~ /\/\.\//) {
	$path2 =~ s|/\./|/|;
    }

    # Squash out //'s in paths
    while ($path1 =~ /\/\//) {
	$path1 =~ s|//|/|;
    }

    while ($path2 =~ /\/\//) {
	$path2 =~ s|//|/|;
    }

    if ($path1 eq $path2) {
	return 1;
    } else {
	return 0;
    }
}

sub printEditRCError {
    print << "EOM";

Error editing the /etc/rc file.  VMware will still install, but dual
configuration will not be supported unless you manually edit the file.
Consult the README file for more information (look for "dualconf").

EOM

}

#
# editRC
#
# Copy the /etc/rc file to /tmp/rc.new and insert the call 
# to /etc/vm.conf, enabling dual-booting of BSD.  This new file
# (/tmp/rc.new) will then be installed over rc.d as one of the actions
#

sub  editRC {
    my @lines;
    my $i, $currentLine = 0, $backtrackLine, $insertAfterLine = -1, $ifDepth=0;
    my $garbage;

    open (OLDRC, "</etc/rc") or die "\n### open: /etc/rc $!.\n";
    open (NEWRC, ">/tmp/rc.new") or die "\n### open: rc.new $!.\n";
    for($currentLine=0; defined($_ = <OLDRC>); $currentLine++){
	
	$lines[$currentLine] = $_;
	
	# OLD INSERTION CASE
	# Get Rid of any old VMware Insertions#

	if(m/\<InstalledByVMware\>/){

	print "Getting rid of old VMware reference in \/etc\/rc\n";

		$currentLine--;  #So the line containing the beginning of the old
	    while(1){
		                 #insertion is overwritten	
		if (!defined ($garbage = <OLDRC>)){
		    &printEditRCError();
		    return;		
		};
		if ($garbage =~ m/\<\/InstalledByVMware\>/){last;}
	    }

	}


	# DONE CASE

	if($insertAfterLine >=0){next;} #once we've inserted the line,
	                                #we're done

	# FOUND START OF NETWORKING

	if((m/ \/etc\/rc.network/  || m/ \/etc\/netstart/) 
		&& !m/^\#/ ){ #match rc.network,not in a comm
	    $insertAfterLine =-2; #error condition:if this isn't changed,die
	    for($backtrackLine = $currentLine-1; $backtrackLine >=0;
		$backtrackLine--){
		my $thisLine = $lines[$backtrackLine];
		#if comment, blank, or within if, move up.
		if($ifDepth > 0 && !$thisLine=~m/^\#/
 			&& !$thisLine=~m/^echo/){
		    
		    if($thisLine =~ m/^if[^a-zA-Z0-9]/ ||
		       $thisLine =~ m/[\t ]if[^a-zA-Z0-9]/){$ifDepth--;}
		    
		    if($thisLine =~ m/^fi[^a-zA-Z0-9]/ ||
		       $thisLine =~ m/[\t ]fi[^a-zA-Z0-9]/){$ifDepth++;}
		    
		    next;
		}
		if($thisLine =~ /^\#/){next;}
		
		#We're here: the current line is the first non-commented
		#line not encased in an if statement before
		#the network call.  Thus we insert our call*after* 
		#the current line.
		
		$insertAfterLine = $backtrackLine;
		last;
	    }
	}

	if($insertAfterLine >=0) {next;} #in case we just set it;
	if($insertAfterLine == -2) {
	    &printEditRCError();
	    return;
	}

	    #ADD TO IFDEPTH CASE
	if(m/^\#/) {next;}
	if(m/^if[^a-zA-Z0-9]/ || m/[\t ]if[^a-zA-Z0-9]/){$ifDepth++;}
	if(m/^fi[^a-zA-Z0-9]/ || m/[\t ]fi[^a-zA-Z0-9]/){
		$ifDepth--;
		if ($ifDepth < 0) {
		    die "\nWeird /etc/rc file: appears to have more fi's than if's!";
		}
	    }
    }
    
      #Output the new rc file

	my $numLines = $#lines+1;

    for($i=0; $i<$numLines; $i++){
	print NEWRC $lines[$i];
	if($i == $insertAfterLine){
	    print NEWRC <<"EOM";	    
### <InstalledByVMware> ### Do not edit this tag!

#Added by the VMware tools to enable dual configurations (one configuration 
#if the system is booted in a VM, another if it is on a real machine.)
#If you remove these lines, you will be locked into the present configuration.
#Also, anything added between the tags in will be overwritten by future
#installations of VMware.

if [ -x /etc/vmware/checkvm ]
then
    if /etc/vmware/checkvm > /dev/null
    then
        if [ -f /etc/rc.dualconf.vm ]; then
            . /etc/rc.dualconf.vm
        fi
    else 
        if [ -f /etc/rc.dualconf.org ]; then
            . /etc/rc.dualconf.org
        fi
    fi   
fi

### </InstalledByVMware> ### Do not edit this tag!
EOM
}
}

#if ($insertAfterLine < 0) die ("\n### Weird /etc/rc file!!!");

close(OLDRC);
close(NEWRC);	  
}


#
# fixConf(newConf, oldConf)
#
# Copy and fixup the rc.conf file for the VM.  Currently just
# turn off the mouse daemon if it is on.
#
sub fixConf {
    my ($newfile, $oldfile) = @_;
    open(OLDF, "<$oldfile") or die "\n### open: $oldfile: $!.\n";
    open(NEWF, ">$newfile") or die "\n### open: $newfile: $!.\n";

    while(defined($_ = <OLDF>)) {
	$line = $_;
	if (/^\s*moused_enable\s*=/) {
	    $line =~ s/YES/NO/;
	    $line =~ s/Yes/No/;
	    $line =~ s/yes/no/;
        }
	print NEWF $line;
    }

    close (OLDF);
    close (NEWF);
}


#
# mergeXconf(newXconf, oldXconf)
#
# Merge the XF86Config information from the old file into the new
# file that we will use when booting in a VM
#
sub mergeXconf {
    my ($newfile, $oldfile) = @_;
    my ($filessection, $kbdsection, $transcribe);
    my ($line, $copy);

    if (not -e $oldfile) {
	my $command = "touch " . $oldfile;
	if (system($command) != 0) {
	    print "The command\n  " . $command . "\n failed: $!.\n";
	    exit(1);
	}
    }

    $filessection = 0;
    $kbdsection = 0;
    $transcribe = 0;

    # Capture current contents before overwriting
    my @inputs = `cat $newfile`;
    open(OLDF, "<$oldfile") or die "\n### open: $oldfile: $!.\n";
    open(NEWF, ">$newfile") or die "\n### open: $newfile: $!.\n";
    while(defined($_ = <OLDF>)) {
	$line = $_;
	if (/^\s*#/) {
	    if ($transcribe == 0) {
		print NEWF $line;
	    }
	} else {
	    if (/^\s*EndSection/) {
		if ($transcribe == 1) {
		    $transcribe = 2;
		}
	    } elsif (/^\s*Section/) {
		s/^\s*Section\s*\(["a-zA-Z]*\).*/$1/;
		if (/Files/) {
		    $filesection = 1;
		    $transcribe = 1;
		} elsif (/Keyboard/) {
		    $kbdsection = 1;
		    $transcribe = 1;
		}
	    }
	}

	if ($transcribe == 1) {
	    print NEWF $line;
	} elsif ($transcribe == 2) {
	    $transcribe = 3;
	    print NEWF $line;
	}
    }

    # Then append our sections that weren't superceded
    $transcribe = 0;
    foreach $line (@inputs) {
	$copy = $line;
	if (not ($copy =~ /^\s*#/)) {
	    if ($copy =~ /^\s*EndSection/) {
		if ($transcribe == 1) {
		    $transcribe = 2;
		}
	    } elsif ($copy =~ /^\s*Section/) {
		$copy =~ s/^\s*Section\s*\(["a-zA-Z]*\).*/$1/;
		if ($copy =~ /Files/) {
		    if ($filesection) {
			$transcribe = 1;
		    }
		} elsif ($copy =~ /Keyboard/) {
		    if ($kbdsection) {
			$transcribe = 1;
		    }
		}
	    }
	}

	if (($transcribe != 1) and ($transcribe != 2)) {
	    print NEWF $line;
	}

	if ($transcribe == 2) {
	    $transcribe = 3;
	}
    }

    close (OLDF);
    close (NEWF);
}


#
# makeLinked(filename)
#
# Convert the given file to a symbolic link to file with the same name
# and the extension ".org".
#
sub makeLinked {
    my ($filename) = @_;
    if (not -e $filename) {
	my $command = "touch " . $filename;
	if (system($command) != 0) {
	    print "The command\n  " . $command . "\n failed: $!.\n";
	    exit(1);
	}
    }

    # First check if we are already set up
    if (-l $filename) {
	my $linkname = readlink($filename);
	my $dirname = substr($filename, 0, rindex($filename, "/"));
	if (&pathCompare($dirname, $linkname, $filename . ".org") or
	    &pathCompare($dirname, $linkname, $filename . ".vm")) {
	    return;
	}
    }
    &install("move", 0, $filename, $filename . ".org", 1);
    &install("link", 0, $filename . ".org", $filename, 0);
}

#
# End of subroutines, start of main procedure
#

&OpenLog;



#
# Try to determine the current screen size
#
my $vm=`./checkvm`;
$screen_mode = "0x0";
if ($vm =~ /good/) {
   $screen_mode = `./checkvm -r`;
   chomp($screen_mode);
}

if ($screen_mode =~ /0x0/) {
   $screen_mode = "\"1600x1200\" \"1280x1024\" \"1152x864\" \"1024x768\" \"800x600\" \"640x480\"";
}

#
# OK, we can fill in the table now
#

# BSD systems are a bit simpler than Linux, since everything is taken care
# of in /etc/rc.conf 
# These values work for freeBSD3.1  Will have to check other BSD's

$xserver_bin = "/usr/X11R6/bin";   #just the directory
$xserver_link = $xserver_bin . "\/X";
$xconfig_file = "/etc/XF86Config"; #full patha and filename

#check for X server binary and config file
 
if ((-e $xserver_bin) && (-e $xserver_link)  && (-e $xconfig_file )){
    print <<"EOM";
Your system appears to have the X server binary, link, and config file in their
standard locations for BSD systems.  You should be able to use the default
locations given below (just hit enter.)
EOM
}else{
    print <<"EOM";
The X server binary directory, link file, and/or config file were not found in
their expected locations ($xserver_bin, $xserver_link, and 
$xconfig_file, respectively.)  Either you do not have X installed,
or X is using non-standard locations for one of these files.

If X is not currently installed and configured, you will need to re-install 
the VMware tools after installing X in order to use the VMware X Server.
If you plan on installing X but have not yet done so, you are advised to
install X before continuing the VMware tools installation.

If X is installed, you need to find its binary directory, link file, and config
file locations, and enter them below:
EOM
}



print("\nDirectory where X server binary is located [$xserver_bin]: "
); 
      chop($reply = <STDIN>);
      $reply =~ s/^\s*//;			# Kill leading whitespace
      $reply =~ s/\s*$//;			# Kill trailing whitespace
if ($reply){$xserver_bin = $reply};

print("\nX file link \(this is a file named \"X\" which is currently a link
to some file in $xserver_bin.\) \[$xserver_link\]: "); 
       chop($reply = <STDIN>);
       $reply =~ s/^\s*//;			# Kill leading whitespace
       $reply =~ s/\s*$//;			# Kill trailing whitespace
if ($reply){$xserver_link = $reply};

print("\nX configuration file [$xconfig_file]: "
      );
      chop($reply = <STDIN>);
      $reply =~ s/^\s*//;			# Kill leading whitespace
      $reply =~ s/\s*$//;			# Kill trailing whitespace
if ($reply){$xconfig_file = $reply};

my @ACTIONS =
(
# ACTION     SOURCE               DESTINATION                   MODE    BACKUP?

 ["copy",    "/tmp/rc.new",       "/etc/rc",                     "0755",   1 ],
 ["dist",    "dualconf.vm.dist",  "/etc/rc.dualconf.vm",         "0755",   1 ],
 ["dist",    "dualconf.org.dist", "/etc/rc.dualconf.org",        "0755",   1 ],
 ["copy",    "checkvm",           "/etc/vmware/checkvm",         "0755",   1 ],
 ["copy",    "XF86_VMware",       "$xserver_bin/XF86_VMware",    "0755",   1 ],
 ["copy",    "vmware-toolbox",    "$xserver_bin/vmware-toolbox", "0755",   0 ],
 ["copy",    "toolbox-wish",      "/etc/vmware/toolbox-wish",    "0755",   0 ],
 ["copy",    "toolbox.tcl",	  "/etc/vmware/toolbox.tcl",	 "0644",   0 ],
 ["copy",    "helpText.tcl",	  "/etc/vmware/helpText.tcl",	 "0644",   0 ],
 ["copy",    "lib/util.tcl",	  "/etc/vmware/lib/util.tcl",	 "0644",   0 ],
 ["copy",    "lib/tabs.tcl",	  "/etc/vmware/lib/tabs.tcl",	 "0644",   0 ],
 ["copy",    "lib/dialog.tcl",	  "/etc/vmware/lib/dialog.tcl",	 "0644",   0 ],
 ["copy",    "lib/help.tcl",	  "/etc/vmware/lib/help.tcl",	 "0644",   0 ],
 ["copy",    "lib/htmlButton.tcl","/etc/vmware/lib/htmlButton.tcl","0644", 0 ],
 ["copy",    "lib/shell.tcl",	  "/etc/vmware/lib/shell.tcl",	 "0644",   0 ],
 ["copy",    "global.help",       "/etc/vmware/global.help",     "0644",   0 ],
 ["copy",    "logo.gif"   ,       "/etc/vmware/logo.gif",        "0644",   0 ],
 ["copy",    "icon.gif"   ,       "/etc/vmware/icon.gif",        "0644",   0 ],
 ["move",    "/tmp/XF86Config.dist",
				  "$xconfig_file" . ".vm",       "0644",   1 ],
 ["link",    "$xserver_bin/XF86_VMware",
			          "$xserver_link" . ".vm",        0755,    1 ],
 );

#
# Tell what we're going to do.
#
print "\nThe following actions will be performed:\n\n";

foreach $d (@ACTIONS) {
    my $action = $$d[0];
    my $source = $$d[1];
    my $dest   = $$d[2];
    my $mode   = $$d[3];
    my $back   = $$d[4];

    print $action eq "link" ? "Link " :  "Copy ";
    print $source . " to " . $dest . "\n";
}

#my $init_links = 0;

print "\nLook at the README file for more information if you are\n";
print "unsure about any of these changes.\n";

print "\nAll existing files, except for the toolbox files, will be backed up"
    . "\n(xxx -> xxx.old.n)."
    . "\n\nOK? [yes] ";

chop($reply = <STDIN>);
$reply =~ s/^\s*//;			# Kill leading whitespace
$reply =~ s/\s*$//;			# Kill trailing whitespace

if ($reply ne "yes" && $reply ne "" && $reply ne "y") {
    exit(0);
}

#
# Before the table driven actions are done, merge the old configuration
# information into the new config files.
#
&editRC();

if (-e $xconfig_file) {
    &install("dist", "0644", "XF86Config.dist", "/tmp/XF86Config.dist", 0);
    &mergeXconf("/tmp/XF86Config.dist", $xconfig_file);
}


foreach $d (@ACTIONS) {
    my $action = $$d[0];
    my $source = $$d[1];
    my $dest   = $$d[2];
    my $mode   = $$d[3];
    my $back   = $$d[4];

    my $fulldest = -d $dest ? $dest . "/" . $source : $dest;

    &install($action, $mode, $source, $fulldest, $back);
}

#
# Set up for the boot scripts
#

#
# Make links for rc.conf.  The files rc.conf.org and rc.conf.vm produced by 
# this script may be altered by the user to implement different behavior
# (e.g. network address) depending on whether the system is running in a VM.
#
&makeLinked("/etc/rc.conf");
&fixConf("/etc/rc.conf.vm", "/etc/rc.conf.org");

#Make links for X

&makeLinked($xconfig_file);
&makeLinked($xserver_link);

# Set up links for VM
if (-x "/etc/vmware/checkvm") {
    my $vm = `/etc/vmware/checkvm`;
    if ($vm =~ /good/) {
	if (-e "/etc/rc.dualconf.vm") {
            system("sh /etc/rc.dualconf.vm");
        }
        system("killall moused");
    } else {
	if (-e "/etc/rc.dualconf.org") {
            system("sh /etc/rc.dualconf.org");
        }
    }
}


# Print out informational messages
#
if (0) { #not needed in BSD; kept for readme purposes
    print << "EOM";

This script was unable to setup links to scripts to run at
init time.  These scripts automatically configure the system
according to whether it is being run in a virtual machine, or
on the real hardware.  There is more information on this topic
in the README file in this directory, if you wish to setup
these scripts to run at init time yourself.
EOM
}

print << "EOM";

Please make sure that glibc is installed.  If it is not,
the X server will fail to run and produce this message:

    "execve failed for /usr/X11R6/bin/XF86_VMware (errno 2)"
EOM

print "\nThe installation has completed.\n";

print "\nFor dual booting this BSD system, the system configuration\n";
print "may differ between the real hardware and the virtual hardware.\n";
print "\nPlease reboot at this time, to complete the system configuration\n";
print "for the virtual machine.  See the README file in this directory\n";
print "for more information about dual booting.\n";

if (not ((-e $xserver_bin) && (-e $xserver_link)  && (-e $xconfig_file ))) { 
print<<"EOM"

It appears that either your X Window system was not fully configured, or you
entered incorrect path information.  If your X window system was not properly
configured, you may have to configure it and then re-install the VMware tools.
If you entered incorrect path information, you can re-install the VMware tools
with new path information at any time.
EOM
}

system($xserver_bin . "/vmware-toolbox > /tmp/toolbox.test 2>&1");
open(TOOLTEST, "</tmp/toolbox.test");
if ($_ = (<TOOLTEST>)){
	if (m/not found/){
print <<"EOM";

It appears that your system does not have tcl and tk properly installed. 
The installation was not affected, but the toolbox will not work until
you have installed tcl/tk, version 8.0 or higher. 
EOM
}
}
close(TOOLTEST);
system("rm /tmp/toolbox.test");

