#!/usr/bin/perl -w
#
# Copyright (C) 1999 VMware, Inc.  All Rights Reserved.
#
# Linux guest OS tools installation
#
# If your copy of perl is not in /usr/bin, please adjust the line above.


die "You must be root to install the VMware Guest OS Tools.\n" unless $> == 0;

print << "EOM";
---------------------------------------
    VMware Guest OS Tools Installer
    Copyright (C) 1999 VMware, Inc.
---------------------------------------

EOM

my ($default_initdir, $modules_file, $xconfig_file, $xserver_link);
my ($screen_mode, $modules_src_file);


#
# makeDirectory(dir, logging)
#
# Make the directory, if it doesn't exist
#
sub makeDirectory {
    my ($path, $logging) = @_;
    my $slashpos;
    my $head;

    if ("$path" eq "") {
	return;
    }

    if (-d $path) {
	return;
    }

    if (-e "$path") {
	die "\n### $path exists but it is not a directory.\n";
    } else {
	$slashpos = rindex($path, "/");
	$head = substr($path, 0, $slashpos);
	&makeDirectory($head, $logging);
	if ($logging) {
	    &Log("directory $path\n");
	}
	mkdir($path, 0755) or die "\n### Cannot create $path: $!.\n";
    }
}


#
# LogFile(filename)
#
# Write filename and modification time to log
#
sub LogFile {
    my $filename = shift;
    my ($dev, $ino, $mode, $nlink, $uid, $gid, $rdev, $size,
	      $atime, $mtime, $ctime) = stat($filename);
    &Log("file $filename $mtime\n");
}


#
# Log(message)
#
# Write message to log
#
sub Log {
  my $string = shift;
  print LINUX_TOOLS_LOG $string;
}


#
# OpenLog
#
# Open the log file, creating it if necessary
#
sub OpenLog {
  &makeDirectory("/etc/vmware", 0);
  open(LINUX_TOOLS_LOG,">/etc/vmware/tools_log") or
     die "\n### Cannot open /etc/vmware/tools_log\n";
}


#
# copydist(mode, src, dest)
#
# Copy the file and set its mode.  If the file has a .dist extension,
# perform keyword substitution first.
#
sub copydist {
  my ($mode, $src, $dst) = @_;

  if (-e $dst) {
      die "\n### File $dst should not exist.\n" .
	  "### Please remove it and try again.\n";
  }

  open(SRC, "<$src") or die "\n### open: $src: $!.\n";
  open(DST, ">$dst") or die "\n### open: $dst: $!.\n";
  my $command = "chmod " . $mode . " " . $dst;
  if (system($command) != 0) {
    print "The command\n  " . $command . "\n failed: $!.\n";
    exit(1);
  }

  if ($src =~ /\.dist$/) {              # Need to make changes to contents
    while(defined($_ = <SRC>)) {
      s/%MODULES_FILE%/$modules_file/g;
      s/%XCONFIG_FILE%/$xconfig_file/g;
      s/%XSERVER_LINK%/$xserver_link/g;
      s/%SCREEN_MODE%/$screen_mode/g;
      print DST $_;
    }
  } else {                              # Just copy
    while(defined($_ = <SRC>)) {
      print DST $_;
    }
  }

  close(SRC);
  close(DST);
}


#
# install(action, mode, src, dst, back)
#
# Install the file in the destination with the given mode.
# Action is one of link/copy/dist.  Dist is like copy, but
# with keyword substitution.
#
sub install {
    my ($action, $mode, $src, $dst, $back) = @_;

    if (-e $dst) {
      if ($back) {
	  if (-e $dst . ".old.3") {
	      system ("rm -f $dst" . ".old.3");
	  }
	TRY:
	  for $i (1..99) {
	      if (! -e $dst . ".old." . $i) {
		  rename($dst, $dst . ".old." . $i);
		  last TRY;
	      }
	  }
       } else {
	  system ("rm -f $dst");
       }
    }

    my $path = substr($dst, 0, rindex($dst, "/"));
    &makeDirectory($path, 1);

    if ($action eq "dist") {
	# Distribute action
	&copydist($mode, $src, $dst);
	&LogFile($dst);
    } elsif ($action eq "copy") {
	# Copy action
	my $command = "cp " . $src . " " . $dst;
	if (system($command) != 0) {
	    print "The command\n  " . $command . "\n failed: $!.\n";
	    exit(1);
	}
	$command = "chmod " . $mode . " " . $dst;
	if (system($command) != 0) {
	    print "The command\n  " . $command . "\n failed: $!.\n";
	    exit(1);
	}
	&LogFile($dst);
    } elsif ($action eq "move") {
	# Move action
	my $command = "mv " . $src . " " . $dst;
	if (system($command) != 0) {
	    print "The command\n  " . $command . "\n failed: $!.\n";
	    exit(1);
	}
	&Log("move $src $dst\n");
    } else {
	# Link action
	my $command = "ln -s " . $src . " " . $dst;
	if (system($command) != 0) {
	    print "The command\n  " . $command . "\n failed: $!.\n";
	    exit(1);
	}
	&Log("link $src $dst\n");
    }
}


#
# pathCompare(dir, path1, path2)
#
# Compare the two paths, and return true if they are identical
# Evaluate the paths with respect to the passed in directory
#
sub pathCompare {
    my ($dir, $path1, $path2) = @_;

    # Prepend directory for relative paths
    $path1 =~ s|^([^/])|$dir/$1|;
    $path2 =~ s|^([^/])|$dir/$1|;

    # Squash out ..'s in paths
    while ($path1 =~ /\/.*\/\.\.\//) {
	$path1 =~ s|/[^/]*/\.\./|/|;
    }

    while ($path2 =~ /\/.*\/\.\.\//) {
	$path2 =~ s|/[^/]*/\.\./|/|;
    }

    # Squash out .'s in paths
    while ($path1 =~ /\/\.\//) {
	$path1 =~ s|/\./|/|;
    }

    while ($path2 =~ /\/\.\//) {
	$path2 =~ s|/\./|/|;
    }

    # Squash out //'s in paths
    while ($path1 =~ /\/\//) {
	$path1 =~ s|//|/|;
    }

    while ($path2 =~ /\/\//) {
	$path2 =~ s|//|/|;
    }

    if ($path1 eq $path2) {
	return 1;
    } else {
	return 0;
    }
}


#
# mergeMods(newXconf, oldXcond)
#
# Merge the modules.conf (conf.modules) information from the old
# file into the new file that we will use when booting in a VM
#
sub mergeMods {
    my ($newfile, $oldfile) = @_;
    my ($inline, $line, $optchk, $count);
    my @inwords;
    my ($ethfound, $sbfound);

    # Capture current contents before overwriting
    my @inputs = `cat $newfile`;
    open(OLDF, "<$oldfile") or die "\n### open: $oldfile: $!.\n";
    open(NEWF, ">$newfile") or die "\n### open: $newfile: $!.\n";

    # Look for matches and selectively replace drivers
    $ethfound = 0;
    $sbfound = 0;
    while(defined($inline = <OLDF>)) {
	$optchk = 0;
	foreach $line (@inputs) {
	    @inwords = split (/\s+/, $line);
	    $count = @inwords;
	    if (defined($inline) and ($count >= 2)) {
		if ($inline =~ /^\s*$inwords[0]\s+$inwords[1]/) {
		    if (($inwords[0] =~ /options/) and ($optchk == 1)) {
			$optchk = 0;
			print NEWF $line;
			$inline = <OLDF>;
		    } elsif ($inwords[0] =~ /alias/) {
			if ($inwords[1] =~ /eth0/) {
			    $ethfound = 1; 
			} elsif (($inwords[1] =~ /sound/) or
				 ($inwords[1] =~ /char-major-14/)) {
			    $sbfound = 1;
			}
			$optchk = 1;
			print NEWF $line;
			$inline = <OLDF>;
		    }
		} elsif ($optchk) {
		    if ($inwords[0] =~ /options/) {
			print NEWF $line;
		    }
		    if ($inline =~ /^\s*options\s+/) {
			$inline = <OLDF>;
		    }
		    $optchk = 0;
		}
	    } 
	}
	if (defined($inline)) {
	    print NEWF $inline;
	}
    }

    # Then pick up any drivers we haven't got yet.
    if (($ethfound == 0) or ($sbfound == 0)) {
	foreach $line (@inputs) {
	    @inwords = split (/\s+/, $line);
	    $count = @inwords;
	    if ($count >= 2) {
		if (($ethfound == 0) and ($inwords[1] =~ /eth0/)) {
		    if ($inwords[0] =~ /options/) {
			$optchk = 0;
		    } else {
			$optchk = 1;
		    }
		    print NEWF $line;
		    $ethfound = 1;
		} elsif (($sbfound == 0) and 
			 (($inwords[1] =~ /sound/) or
			  ($inwords[1] =~ /char-major-14/)) ) {
		    if ($inwords[0] =~ /options/) {
			$optchk = 0;
		    } else {
			$optchk = 1;
		    }
		    print NEWF $line;
		    $sbfound = 1;
		} elsif ($optchk) {
		    if ($inwords[0] =~ /options/) {
			print NEWF $line;
		    }
		    $optchk = 0;
		}
	    }
	}
    }

    close (OLDF);
    close (NEWF);
}


#
# mergeXconf(newXconf, oldXconf)
#
# Merge the XF86Config information from the old file into the new
# file that we will use when booting in a VM
#
sub mergeXconf {
    my ($newfile, $oldfile) = @_;
    my ($filessection, $kbdsection, $transcribe);
    my ($line, $copy);

    $filessection = 0;
    $kbdsection = 0;
    $transcribe = 0;

    # Capture current contents before overwriting
    my @inputs = `cat $newfile`;
    open(OLDF, "<$oldfile") or die "\n### open: $oldfile: $!.\n";
    open(NEWF, ">$newfile") or die "\n### open: $newfile: $!.\n";
    while(defined($_ = <OLDF>)) {
	$line = $_;
	if (/^\s*#/) {
	    if ($transcribe == 0) {
		print NEWF $line;
	    }
	} else {
	    if (/^\s*EndSection/) {
		if ($transcribe == 1) {
		    $transcribe = 2;
		}
	    } elsif (/^\s*Section/) {
		s/^\s*Section\s*\(["a-zA-Z]*\).*/$1/;
		if (/Files/) {
		    $filesection = 1;
		    $transcribe = 1;
		} elsif (/Keyboard/) {
		    $kbdsection = 1;
		    $transcribe = 1;
		}
	    }
	}

	if ($transcribe == 1) {
	    print NEWF $line;
	} elsif ($transcribe == 2) {
	    $transcribe = 3;
	    print NEWF $line;
	}
    }

    # Then append our sections that weren't superceded
    $transcribe = 0;
    foreach $line (@inputs) {
	$copy = $line;
	if (not ($copy =~ /^\s*#/)) {
	    if ($copy =~ /^\s*EndSection/) {
		if ($transcribe == 1) {
		    $transcribe = 2;
		}
	    } elsif ($copy =~ /^\s*Section/) {
		$copy =~ s/^\s*Section\s*\(["a-zA-Z]*\).*/$1/;
		if ($copy =~ /Files/) {
		    if ($filesection) {
			$transcribe = 1;
		    }
		} elsif ($copy =~ /Keyboard/) {
		    if ($kbdsection) {
			$transcribe = 1;
		    }
		}
	    }
	}

	if (($transcribe != 1) and ($transcribe != 2)) {
	    print NEWF $line;
	}

	if ($transcribe == 2) {
	    $transcribe = 3;
	}
    }

    close (OLDF);
    close (NEWF);
}


#
# makeLinked(filename)
#
# Convert the given file to a symbolic link to file with the same name
# and the extension ".org".
#
sub makeLinked {
    my ($filename) = @_;
    if (not -e $filename) {
	my $command = "touch " . $filename;
	if (system($command) != 0) {
	    print "The command\n  " . $command . "\n failed: $!.\n";
	    exit(1);
	}
    }

    # First check if we are already set up
    if (-l $filename) {
	my $linkname = readlink($filename);
	my $dirname = substr($filename, 0, rindex($filename, "/"));
	if (&pathCompare($dirname, $linkname, $filename . ".org") or
	    &pathCompare($dirname, $linkname, $filename . ".vm")) {
	    return;
	}
    }
    &install("move", 0, $filename, $filename . ".org", 1);
    &install("link", 0, $filename . ".org", $filename, 0);
}


#
# End of subroutines, start of main procedure
#
&OpenLog;

#
# Find what system we're on.
#
my ($system, $reply);
my @SYSTEMS = ("redhat", "suse", "caldera", "turbolinux", "other");
my $issue = "/etc/issue";

if (-e $issue) {
    $system = $#ARGV >= 0 && $ARGV[0] ||
	      `grep -i "red *hat" $issue` && "redhat" ||
	      `grep -i "turbolinux" $issue` && "turbolinux" ||
	      `grep -i "suse" $issue` && "suse" ||
	      `grep -i "s\.u\.s\.e" $issue` && "suse" ||
	      `grep -i "caldera" $issue` && "caldera" ||
	      "unknown";
} else {
    $system = $#ARGV >= 0 && $ARGV[0] ||
	      "unknown";
}

if (scalar(grep(/^\Q$system\E$/, @SYSTEMS)) == 0) {
    print $system eq "unknown" ?
	  "Can't determine your Linux installation.\n" :
	  "Don't know this kind of Linux installation: \"$system\".\n";
    print "Please rerun this script in one of these ways:\n";
    foreach $s (@SYSTEMS) {
	print "  $0 $s\n"
    }
    exit(1);
}
if ($#ARGV < 0 || $system ne $ARGV[0]) {
    print "You seem to have a" .
	  ($system =~ m/^[aeiou]/ ? "n" : "") .
	  " $system installation.\n";
}
&Log("system $system\n");


#
# Set global system dependent configuration variables
#
$default_initdir =
    defined($ENV{INITDIR})? $ENV{INITDIR} :
    -d "/etc/init.d" ? "/etc" :
    -d "/etc/rc.d" ? "/etc/rc.d" : "/etc";

$modules_file = -e "/etc/modules.conf" ? "/etc/modules.conf" :
		"/etc/conf.modules";

$xconfig_file = "NULL";
$xserver_link = "NULL";
$xserver_bin = "/usr/X11R6/bin";

if ($system eq "redhat") {
    $xconfig_file = "/etc/X11/XF86Config";
    $xserver_link = "/etc/X11/X";
}

if ($system eq "turbolinux") {
    $xconfig_file = "/etc/X11/XF86Config";
    $xserver_link = "/etc/X11/X";
}

if ($system eq "suse") {
    $xconfig_file = "/etc/XF86Config";
    $xserver_link = "/var/X11R6/bin/X";
}

if ($system eq "caldera") {
    $xconfig_file = "/etc/XF86Config";
    $xserver_link = "/usr/X11R6/bin/X";
}

#
# If the system is unknown, try to guess the locations
#
if ($xconfig_file eq "NULL") {
    $xconfig_file = -e "/etc/X11/XF86Config" ? "/etc/X11/XF86Config" :
	-e "/etc/XF86Config" ? "/etc/XF86Config" :
	"/etc/X11/XF86Config";
}

if ($xserver_link eq "NULL") {
    if (-e "/etc/X11/Xserver") {
	my @servlist = `cat /etc/X11/Xserver`;
	my ($server, $rest) = @servlist;
	chomp ($server);
	if (-e $server) {
	    $xserver_link = $server;
	}
    }
}

if ($xserver_link eq "NULL") {
    $xserver_link = -e "/etc/X11/X" ? "/etc/X11/X" :
		    -e "/var/X11R6/bin/X" ? "/var/X11R6/bin/X" :
		    -e "/usr/X11R6/bin/X" ? "/usr/X11R6/bin/X" :
		    "/etc/X11/X";
}

#
# We should use the pcnet32 driver, if it is available, otherwise
# use the lance network driver.
#
$modules_src_file = "conf.modules.vm";

my $drivers = `modprobe -l -t net`;
if ($drivers =~ /pcnet32/) {
  $modules_src_file = "conf.modules.vm";
} elsif ($drivers =~ /lance/) {
  $modules_src_file = "conf.modules.vm_lance";
}

#
# Try to determine the current screen size
#
my $vm=`./checkvm`;
$screen_mode = "0x0";
if ($vm =~ /good/) {
   $screen_mode = `./checkvm -r`;
   chomp($screen_mode);
}

if ($screen_mode =~ /0x0/) {
   $screen_mode = "\"1600x1200\" \"1280x1024\" \"1152x864\" \"1024x768\" \"800x600\" \"640x480\"";
}

#
# OK, we can fill in the table now
#
my @ACTIONS =
(
# ACTION     SOURCE               DESTINATION                   MODE    BACKUP?

 ["dist",    "dualconf.vm.dist",  "/etc/vmware/dualconf.vm",     "0755",   1 ],
 ["dist",    "dualconf.org.dist", "/etc/vmware/dualconf.org",    "0755",   1 ],
 ["copy",    "checkvm",           "/etc/vmware/checkvm",         "0755",   1 ],
 ["copy",    "XF86_VMware",       "$xserver_bin/XF86_VMware",    "0755",   1 ],
 ["copy",    "vmware-toolbox",    "$xserver_bin/vmware-toolbox", "0755",   1 ],
 ["move",    "/tmp/XF86Config.dist",
				  "$xconfig_file" . ".vm",       "0644",   1 ],
 ["link",    "$xserver_bin/XF86_VMware",
			          "$xserver_link" . ".vm",        0,       1 ],
 ["copy",    "dualconf",          "$default_initdir/init.d/dualconf",
								 "0755",   1 ],
 ["move",    "/tmp/" . "$modules_src_file",
				  "$modules_file" . ".vm",       "0644",   1 ],
 );



#
# Tell what we're going to do.
#
print "The following actions will be performed:\n\n";

foreach $d (@ACTIONS) {
    my $action = $$d[0];
    my $source = $$d[1];
    my $dest   = $$d[2];
    my $mode   = $$d[3];
    my $back   = $$d[4];

    print $action eq "link" ? "Link " :  "Copy ";
    print $source . " to " . $dest . "\n";
}

my $init_links = 0;
if (-d "$default_initdir/init.d") {
    print "\nAlso, links to " . $default_initdir . "/init.d/dualconf";
    print " will be placed in " . $default_initdir . '/rc*.d' . ".\n";
    $init_links = 1;
}

print "\nLook at the README file for more information if you are\n";
print "unsure about any of these changes.\n";

print "\nAll existing files and links will be backed up (xxx -> xxx.old.n)."
    . "\n\nOK? [yes] ";

chop($reply = <STDIN>);
$reply =~ s/^\s*//;			# Kill leading whitespace
$reply =~ s/\s*$//;			# Kill trailing whitespace

if ($reply ne "yes" && $reply ne "" && $reply ne "y") {
    exit(0);
}

#
# Before the table driven actions are done, merge the old configuration
# information into the new config files.
#
if (-e $modules_file ) {
    &install("copy", "0644", $modules_src_file,
			     "/tmp/" . "$modules_src_file", 0);
    &mergeMods("/tmp/" . "$modules_src_file", $modules_file);
}

if (-e $xconfig_file) {
    &install("dist", "0644", "XF86Config.dist", "/tmp/XF86Config.dist", 0);
    &mergeXconf("/tmp/XF86Config.dist", $xconfig_file);
}

foreach $d (@ACTIONS) {
    my $action = $$d[0];
    my $source = $$d[1];
    my $dest   = $$d[2];
    my $mode   = $$d[3];
    my $back   = $$d[4];

    my $fulldest = -d $dest ? $dest . "/" . $source : $dest;

    &install($action, $mode, $source, $fulldest, $back);
}

#
# Set up for the boot scripts
#
&makeLinked($modules_file);
&makeLinked($xconfig_file);
&makeLinked($xserver_link);

#
# Find the existing network scripts, and link in the dualconf
# script in the same directories, with the same number.
#
if ($init_links) {
    my ($line, $source, $newlink);
    $source = $default_initdir . "/init.d/dualconf";

    @nscripts = `find $default_initdir -follow -name 'S*network' -print`;

    foreach $line (@nscripts) {
	($newlink = $line) =~ s/S(.*)network/S$1dualconf/;
	chomp ($newlink);
	&install("link", "0755", $source, $newlink, 0);
    }
}

#
# Run the init time script to configure for the current environment
#
system("sh " . $default_initdir . "/init.d/dualconf");

#
# Print out informational messages
#
if (not $init_links) {
    print << "EOM";

This script was unable to setup links to scripts to run at
init time.  These scripts automatically configure the system
according to whether it is being run in a virtual machine, or
on the real hardware.  There is more information on this topic
in the README file in this directory, if you wish to setup
these scripts to run at init time yourself.
EOM
}

if (($system ne "redhat") and ($system ne "suse") and
    ($system ne "turbolinux")) {
    print << "EOM";

Please make sure that glibc is installed.  If it is not,
the X server will fail to run and produce this message:

    "execve failed for /usr/X11R6/bin/XF86_VMware (errno 2)"
EOM
}

print "\nThe installation has completed.\n";

print "\nFor dual booting this Linux system, the system configuration\n";
print "may differ between the real hardware and the virtual hardware.\n";
print "\nPlease reboot at this time, to complete the system configuration\n";
print "for the virtual machine.  See the README file in this directory\n";
print "for more information about dual booting.\n";

